from __future__ import annotations

"""
FastAPI Extensions: OAuth2 Manager
"""

import abc
from lzl.types import eproperty
from lzl.logging import logger
from lzl import load
from lzo.utils.helpers.base import create_unique_secret_key
from starlette.authentication import AuthenticationError
from lzl.ext.fast.types.security import Authorization
from urllib.parse import urljoin
from . import errors
# from .types.security import HTTPBase, HTTPAuthorizationCredentials, Authorization
from typing import Optional, Dict, Any, List, Union, Type, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from fastapi import FastAPI
    from .config import OAuth2Settings
    from .providers import ProviderClientT
    from .types import Request
    from .types.user import OAuth2User
    from .types.token import OAuth2Token
    from .types.security import HTTPBase, HTTPAuthorizationCredentials
else:
    from fastapi import Request

if load.TYPE_CHECKING:
    import jwt
else:
    jwt = load.LazyLoad("jwt", install_missing=True, install_options={'package': 'pyjwt'})

_DefaultRoutes: Dict[str, Any] = {
    'login': '/login',
    'start': '/start',
    'callback': '/callback',
    'logout': '/logout',
    'user_info': '/userinfo',
    'default_redirect': '/docs',
}

class OAuth2Manager(abc.ABC):
    """
    The OAuth2 Manager
    """
    _extra: Dict[str, Any] = {}

    def __init__(
        self,
        settings: 'OAuth2Settings',
        **kwargs,
    ) -> None:
        """
        Initializes the OAuth2 Manager
        """
        self.settings = settings
        self.providers: Dict[str, 'ProviderClientT'] = {}
        self.jwt_issuers: Dict[str, str] = {} # A proxy for jwt_issuer_url and provider name
        self.default_provider: Optional[str] = None
        self.routes: Dict[str, Any] = _DefaultRoutes.copy()
    
    @property
    def user_class(self) -> Type['OAuth2User']:
        """
        Returns the User Class
        """
        if 'user_class' not in self._extra:
            from .types.user import OAuth2User
            self._extra['user_class'] = OAuth2User
        return self._extra['user_class']
    
    @user_class.setter
    def user_class(self, value: Union[Type['OAuth2User'], str]) -> None:
        """
        Sets the User Class
        """
        if isinstance(value, str): value = load.lazy_import(value)
        self._extra['user_class'] = value

    @property
    def token_class(self) -> Type['OAuth2Token']:
        """
        Returns the Token Class
        """
        if 'token_class' not in self._extra:
            from .types.token import OAuth2Token
            self._extra['token_class'] = OAuth2Token
        return self._extra['token_class']
    
    @token_class.setter
    def token_class(self, value: Union[Type['OAuth2Token'], str]) -> None:
        """
        Sets the Token Class
        """
        if isinstance(value, str): value = load.lazy_import(value)
        self._extra['token_class'] = value
        # TODO set the downstream classes


    @property
    def provider_pattern(self) -> str:
        """
        Returns the provider pattern

        '^(google|kinde|auth0)$'
        """
        if 'provider_pattern' not in self._extra:
            base_pattern = '|'.join(list(self.providers.keys()))
            self._extra['provider_pattern'] = f'^({base_pattern})$'
        return self._extra['provider_pattern']

    @property
    def app(self) -> 'FastAPI':
        """
        Returns the app
        """
        return self.settings.app
    
    def load_providers(
        self, 
        providers: List[Dict[str, Any]], 
        domains: Optional[List[str]] = None
    ) -> None:
        """
        Loads the providers
        """
        from .providers import get_provider_classes
        for config in providers:
            provider = config['name']
            enabled = config.pop('enabled', True)
            if not enabled: continue
            is_default = config.pop('default', False)
            _config_cls, _client_cls = get_provider_classes(
                provider,
                config_class = config.pop('config_class', None),
                client_class = config.pop('client_class', None),
            )
            _config = _config_cls(domains = domains, **config)
            self.providers[provider] = _client_cls(
                config = _config,
            )
            if _config.jwt_issuer_url: self.jwt_issuers[_config.jwt_issuer_url] = provider
            if is_default: self.default_provider = provider
        if not self.default_provider: self.default_provider = list(self.providers.keys())[0]
        if not self.settings.temp_data.has_logged('providers_loaded'):
            logger.info(f'Loaded |g|{len(self.providers)}|e| OAuth2 Providers: |y|{list(self.providers.keys())}|e| - Default Provider: |g|{self.default_provider}|e|', colored = True)

    def get_oauth_provider(self, provider: str) -> 'ProviderClientT':
        """
        Returns the OAuth2 provider
        """
        if provider not in self.providers:
            raise errors.InvalidProviderError(detail=f'Invalid Provider: {provider}')
        return self.providers[provider]

    def get_jwt_token_from_request(self, request: 'Request') -> Optional[str]:
        """
        Gets the JWT Token from the Request
        """
        if not request.headers.get(self.settings.header): return None
        auth = request.headers.get(self.settings.header)
        scheme, token = auth.split(' ', 1)
        return None if scheme.lower() != self.settings.scheme.lower() else token

    async def get_provider_for_jwt(
        self,
        token: str,
        request: 'Request',
    ) -> 'ProviderClientT':
        """
        Resolves the provider for the jwt
        """
        decoded: Dict[str, Union[str, List[str]]] = jwt.decode(
            token,
            algorithms = ['RS256'],
            options={
                'verify_signature': False,
            }
        )
        if decoded.get('iss') not in self.jwt_issuers:
            raise AuthenticationError(detail=f'Invalid Provider: {decoded.get("iss")}')
        return self.providers[self.jwt_issuers[decoded.get('iss')]]
    

    def parse_token_payload(
        self,
        token: Dict[str, Any],
        provider: 'ProviderClientT',
    ) -> Optional['OAuth2Token']:
        """
        Parses the token payload
        """
        auth_token = self.token_class(**token, provider = provider.name)
        if auth_token.access_token:
            if decoded := provider.parse_jwt_token(auth_token.access_token, kind = 'access_token'):
                auth_token.set_access_token(decoded)
        if auth_token.id_token:
            if decoded := provider.parse_jwt_token(auth_token.id_token, kind = 'id_token'):
                auth_token.set_id_token(decoded)
        if auth_token.refresh_token and auth_token.identity:
            provider.refresh_data[auth_token.identity] = auth_token.refresh_token
        return auth_token
    
    async def arefresh_oauth_token(
        self,
        oauth_token: 'OAuth2Token',
        provider: 'ProviderClientT',
        refresh_token: Optional[str] = None,
        **kwargs,
    ) -> Optional['OAuth2Token']:
        """
        Refreshes the oauth token
        """
        if not refresh_token:
            refresh_token = oauth_token.refresh_token or \
                await provider.refresh_data.aget(oauth_token.identity)
        try:
            refreshed_token = await provider.arefresh_token(refresh_token = refresh_token, **kwargs)
            logger.info(f'Refreshed OAuth Token for |g|{oauth_token.identity}|e|', colored = True, prefix = provider.name)
        
        except Exception as e:
            logger.warning(f'Error refreshing token: {e} - Deleting {oauth_token.identity}', prefix = provider.name)
            await provider.data.adelete(oauth_token.identity)
            await provider.refresh_data.adelete(oauth_token.identity)
            return None
        
        refreshed_token = self.parse_token_payload(refreshed_token, provider)
        if not refreshed_token.refresh_token: refreshed_token.refresh_token = refresh_token
        await provider.data.aset(refreshed_token.identity, refreshed_token)
        return refreshed_token
    
    async def aretrieve_token_from_token_payload(
        self,
        token: str,
        token_payload: Dict[str, Any],
        provider: 'ProviderClientT',
        **kwargs,
    ) -> Optional['OAuth2Token']:
        """
        Retrieves the stored token from the token payload
        """
        if not token_payload.get('sub'): 
            logger.info(f'No Subject (sub) in Token: {token_payload}', colored = True, prefix = provider.name)
            raise errors.InvalidTokenError(detail = 'No Subject (sub) in Token')
        
        identity = token_payload.get('sub')
        if await provider.data.acontains(identity):
            # This is likely a token that was issued by this API
            stored_token = await provider.data.aget(identity)
            stored_token.merge(token_payload)
            return stored_token

        # This is likely a token that was issued by another API
        # and thus it isn't stored in the provider's data
        built_token_payload = await provider.abuild_oauth_token(token, token_payload)
        if built_token_payload is None: return None
        built_token = self.token_class(**built_token_payload)
        await provider.data.aset(identity, built_token)
        return built_token

    
    async def authenticate_oauth_token(
        self,
        token: str,
        request: 'Request',
        from_session: Optional[bool] = None,
    ) -> Tuple['OAuth2User', Optional['OAuth2Token']]:
        """
        Authenticates the OAuth2 Token
        """
        provider = await self.get_provider_for_jwt(token, request)
        token_payload = provider.parse_authorization_token(token, verify_exp = False)
        oauth_token = await self.aretrieve_token_from_token_payload(token, token_payload, provider = provider)
        if oauth_token is None:
            return None, self.user_class(None, None, self)
        if oauth_token.is_expired:
            logger.info(f'Refreshing Token: {oauth_token}')
            # Try to refresh the token
            refreshed_token = await self.arefresh_oauth_token(oauth_token, provider = provider)
            if refreshed_token is None: 
                # TODO: handle refresh failure
                if from_session: 
                    _ = request.session.pop('_oauth', None)
                    return None
                raise errors.InvalidSignatureError('Expired Token - Unable to Retrieve Refresh Token')
            oauth_token = refreshed_token
            request.session['_oauth'] = oauth_token.value
        return oauth_token, self.user_class(oauth_token, provider, self)

    

    """
    FastAPI Routes
    """

    def fetch_app_endpoint_from_request(self, request: Request) -> str:
        """
        Fetches the app endpoint from the request
        """
        if request.session.get('app_endpoint'):
            return request.session['app_endpoint']
        headers = dict(request.headers.items())
        if headers.get(':authority:'):
            return headers[':scheme:'] + '://' + headers[':authority:']
        if headers.get('x-forwarded-host'):
            if headers.get('x-forwaded-scheme'):
                return headers['x-forwaded-scheme'] + '://' + headers['x-forwarded-host']
            if self.settings.in_k8s:
                return 'https://' + headers['x-forwarded-host']
            forwaded_port = headers['x-forwarded-port']
            if '443' in forwaded_port: return 'https://' + headers['x-forwarded-host']
            return 'http://' + headers['x-forwarded-host']
        if headers.get('host'):
            scheme = 'https' if self.settings.in_k8s else 'http'
            return f'{scheme}://' + headers['host']
        raise ValueError('Unable to determine app endpoint')

    def mount_oauth_routes(
        self,
        app: Optional['FastAPI'] = None,
        include_in_schema: Optional[bool] = None,
        **kwargs,
    ):
        """
        Mounts the OAuth2 Routes
        """
        if app is not None: self.settings.app = app
        elif self.settings.app is not None: app = self.settings.app
        else: raise ValueError('app must be provided')

        from fastapi import APIRouter,  Path, Query
        from fastapi.responses import RedirectResponse
        from lzl.ext.fast.types.responses import PrettyJSONResponse
        
        router = APIRouter(
            include_in_schema = include_in_schema, 
            default_response_class = PrettyJSONResponse
        )

        @router.get(self.routes['login'])
        async def login(
            request: Request,
            auth: Authorization,
            redirect: Optional[str] = Query(None, description='Redirect URL'),
        ):
            """
            Login Endpoint
            """
            headers = {}
            if request.user.is_authenticated: 
                headers[self.settings.header] = f'{self.settings.scheme} {request.auth.value}'
                if redirect: return RedirectResponse(url = redirect, headers = headers)
                if redirect := request.session.pop('oauth2_redirect', None):
                    return RedirectResponse(url = redirect, headers = headers)
                if self.routes.get('default_redirect'):
                    return RedirectResponse(url = self.routes['default_redirect'], headers = headers)
                return {'login': 'success'}
            provider_url = self.routes['start'] + "/" + self.default_provider
            return RedirectResponse(url = provider_url)


        @router.get(self.routes['start'] + "/{provider}")
        async def start_with_provider(
            request: Request,
            provider: str = Path(..., description='OAuth2 Provider', pattern = self.provider_pattern),
            redirect: Optional[str] = Query(None, description='Redirect URL'),
            scope: Optional[str] = Query(None, description='Scope'),
        ):
            """
            Authorize Endpoint with an OAuth2 Provider
            """
            if redirect: request.session['oauth2_redirect'] = redirect
            state_id = create_unique_secret_key()
            request.session['state_id'] = state_id
            client = self.get_oauth_provider(provider)
            provider_cb_path = self.routes['callback'] + "/" + provider
            provider_redirect_url = urljoin(self.fetch_app_endpoint_from_request(request), provider_cb_path)
            auth_redirect_url = client.get_auth_url(
                redirect_url = provider_redirect_url,
                scope = scope,
                state = state_id,
            )
            return RedirectResponse(auth_redirect_url)
        
        @router.get(self.routes['callback'] + "/{provider}")
        async def callback_with_provider(
            request: Request,
            provider: str = Path(..., description='OAuth2 Provider', pattern = self.provider_pattern),
            code: str = Query(..., description='Authorization Code'),
            state: Optional[str] = Query(None, description='State'),
        ):
            """
            The Provider Callback Endpoint
            """
            client = self.get_oauth_provider(provider)
            provider_cb_path = self.routes['callback'] + "/" + provider
            provider_redirect_url = urljoin(self.fetch_app_endpoint_from_request(request), provider_cb_path)
            try:
                token = await client.afetch_token(
                    code = code,
                    redirect_url = provider_redirect_url,
                    request = request,
                )
            except Exception as e:
                logger.error(f'Error fetching token: {e}')
                raise errors.InvalidAuthorizationError(error = e) from e
            logger.info(f'Fetched Token: |g|{token}|e|', colored = True, prefix = provider)
            oauth_token = self.parse_token_payload(token, client)
            logger.info(f'Fetched Token: |g|{oauth_token}|e|', colored = True, prefix = provider)
            await client.data.aset(oauth_token.identity, oauth_token)
            request.session['_oauth'] = oauth_token.value
            redirect = request.session.get('oauth2_redirect')
            if not redirect and self.routes.get('default_redirect'): redirect = self.routes['default_redirect']
            headers = {
                self.settings.header: f'{self.settings.scheme} {oauth_token.value}'
            }
            if redirect: response = RedirectResponse(url = redirect, headers = headers)
            else: response = PrettyJSONResponse({'login': 'success'}, headers = headers)
            return response


        @router.get(self.routes['logout'])
        async def logout(
            request: Request,
            redirect: Optional[str] = Query(None, description='Redirect URL'),
        ):
            """
            Logout Endpoint
            """
            if request.user.is_authenticated:
                await request.user.alogout()
            if not redirect: redirect = request.session.get('oauth2_redirect')
            request.session.clear()
            if redirect: return RedirectResponse(url = redirect)
            if self.routes.get('default_redirect'):
                return RedirectResponse(url = self.routes['default_redirect'])
            return {'logout': 'success'}
            
        @router.get(self.routes['user_info'])
        async def user_info(
            request: Request,
        ):
            """
            User Info Endpoint
            """
            if not request.user.is_authenticated: return {'user_info': 'not_authenticated'}
            return PrettyJSONResponse(
                content = {
                    'authenticated': True,
                    'authorization': request.auth.value,
                    'user': {
                        'id': request.user.identity,
                        'display_name': request.user.display_name,
                        'email': request.user.email,
                        'roles': request.auth.roles,
                        'scopes': request.auth.scopes,
                        'provider': request.auth.provider,
                    }
                }
            )


        app.include_router(router, tags = ['OAuth2'])

    @classmethod
    def load(
        cls, 
        app: Optional['FastAPI'] = None, 
        **kwargs
    ) -> 'OAuth2Manager':
        """
        Loads the OAuth2 Manager
        """
        from .config import OAuth2Settings
        settings = OAuth2Settings()
        if app is not None: settings.app = app
        if kwargs: settings.update_config(**kwargs)
        return cls(settings = settings)



from __future__ import annotations

"""This module is intended to take the result from EXPLAIN QUERY PLAN and
format in a more human-readable way.
"""

import io # type: ignore
from typing import Dict, List, Optional, TYPE_CHECKING
from .result import ResultItem
from dataclasses import dataclass


if TYPE_CHECKING:
    from rqdb.connection import SyncWritableIO
    from rqdb.async_connection import AsyncWritableIO


@dataclass
class ExplainQueryNode:
    id: int
    detail: str
    parent: Optional["ExplainQueryNode"]
    children: List["ExplainQueryNode"]


@dataclass
class ExplainQueryPlan:
    roots: List[ExplainQueryNode]
    largest_id: int


def parse_explain_query_plan(result: ResultItem) -> ExplainQueryPlan:
    """Parses the given result from EXPLAIN QUERY PLAN into the
    tree structure it represents.
    """
    assert result.results, "Result must have at least one row."
    assert len(result.results[0]) == 4, "Result must have four columns."

    roots: List[ExplainQueryNode] = []
    node_by_id: Dict[int, ExplainQueryNode] = {}
    largest_id = 0

    for row in result.results:
        row_id = row[0]
        detail = row[3]
        parent_id = row[1]

        assert isinstance(row_id, int), "Row ID must be an integer."
        assert isinstance(parent_id, int), "Parent ID must be an integer."
        assert isinstance(detail, str), "Detail must be a string."

        node = ExplainQueryNode(
            id=row_id,
            detail=detail,
            parent=None,
            children=[],
        )
        node_by_id[row_id] = node
        largest_id = max(largest_id, row_id)

        if parent_id == 0:
            roots.append(node)
        else:
            parent_node = node_by_id[parent_id]
            node.parent = parent_node
            parent_node.children.append(node)
    return ExplainQueryPlan(roots=roots, largest_id=largest_id)


def write_explain_query_plan(
    eqp: ExplainQueryPlan,
    out: "SyncWritableIO",
    *,
    indent: int = 3,
    include_raw: bool = False,
) -> None:
    """Given the raw result from EXPLAIN QUERY PLAN, formats it into a
    tree format similar to eqp in the sqlite3 shell.

    Arguments:
        eqp (ExplainQueryPlan): The parsed result from
            parse_explain_query_plan.
        out (SyncWritableIO): The output stream to write to. A buffered
            writer is strongly recommended.
        indent (int): The number of spaces to indent when indicating
            that a line is nested. Must be at least 1
        include_raw (bool): If true, each line is prepended with the
            raw row from the result that was used to generate it.
    """
    assert indent > 0, "Indent must be at least 1."
    largest_id_length = len(str(eqp.largest_id))
    for node in eqp.roots:
        write_explain_query_node(
            node,
            out,
            indent=indent,
            include_raw=include_raw,
            largest_id_length=largest_id_length,
        )


async def async_write_explain_query_plan(
    eqp: ExplainQueryPlan,
    out: "AsyncWritableIO",
    *,
    indent: int = 3,
    include_raw: bool = False,
) -> None:
    """Given the raw result from EXPLAIN QUERY PLAN, formats it into a
    tree format similar to eqp in the sqlite3 shell.

    Arguments:
        eqp (ExplainQueryPlan): The parsed result from
            parse_explain_query_plan.
        out (AsyncWritableIO): The output stream to write to. A buffered
            writer is strongly recommended.
        indent (int): The number of spaces to indent when indicating
            that a line is nested. Must be at least 1
        include_raw (bool): If true, each line is prepended with the
            raw row from the result that was used to generate it.
    """
    assert indent > 0, "Indent must be at least 1."
    largest_id_length = len(str(eqp.largest_id))
    for node in eqp.roots:
        await async_write_explain_query_node(
            node,
            out,
            indent=indent,
            include_raw=include_raw,
            largest_id_length=largest_id_length,
        )


def write_explain_query_node(
    node: ExplainQueryNode,
    out: "SyncWritableIO",
    *,
    level: int = 0,
    indent: int = 3,
    include_raw: bool = False,
    largest_id_length: int = 0,
) -> None:
    """Writes a single explain query node to the given output stream.

    Args:
        node (ExplainQueryNode): The node to write.
        out (SyncWritableIO): The output stream to write to.
        level (int): The current level of indentation, where 0 is no
            indentation, 1 is one level of indentation, and so on.
        indent (int): The number of spaces to indent per level
            of indentation.
        include_raw (bool): If true, each line is prepended with the
            raw row from the result that was used to generate it.
        largest_id_length (int): The length when printed in decimal
            of the largest id in the result, used for padding.
    """
    if include_raw:
        parent_id = node.parent.id if node.parent else 0
        out.write(
            f"[id: {node.id:>{largest_id_length}}, par: {parent_id:>{largest_id_length}}] ".encode(
                "ascii"
            )
        )

    for _ in range(level - 1):
        for _ in range(indent):
            out.write(b" ")

    if level > 0:
        for _ in range(indent - 1):
            out.write(b" ")

    if level > 0:
        out.write(b"|")

    for _ in range(indent - 1):
        out.write(b"-")

    out.write(node.detail.encode("utf-8"))
    out.write(b"\n")

    for child in node.children:
        write_explain_query_node(
            child,
            out,
            level=level + 1,
            indent=indent,
            include_raw=include_raw,
            largest_id_length=largest_id_length,
        )


async def async_write_explain_query_node(
    node: ExplainQueryNode,
    out: "AsyncWritableIO",
    *,
    level: int = 0,
    indent: int = 3,
    include_raw: bool = False,
    largest_id_length: int = 0,
) -> None:
    """Writes a single explain query node to the given output stream.

    Args:
        node (ExplainQueryNode): The node to write.
        out (SyncWritableIO): The output stream to write to.
        level (int): The current level of indentation, where 0 is no
            indentation, 1 is one level of indentation, and so on.
        indent (int): The number of spaces to indent per level
            of indentation.
        include_raw (bool): If true, each line is prepended with the
            raw row from the result that was used to generate it.
        largest_id_length (int): The length when printed in decimal
            of the largest id in the result, used for padding.
    """
    if include_raw:
        parent_id = node.parent.id if node.parent else 0
        await out.write(
            f"[id: {node.id:>{largest_id_length}}, par: {parent_id:>{largest_id_length}}] ".encode(
                "ascii"
            )
        )

    await out.write(b" " * indent * max(level - 1, 0))

    if level > 0:
        await out.write(b" " * (indent - 1))

    if level > 0:
        await out.write(b"|")

    await out.write(b"-" * (indent - 1))

    await out.write(node.detail.encode("utf-8"))
    await out.write(b"\n")

    for child in node.children:
        await async_write_explain_query_node(
            child,
            out,
            level=level + 1,
            indent=indent,
            include_raw=include_raw,
            largest_id_length=largest_id_length,
        )


def format_explain_query_plan_result(
    result: ResultItem, *, indent: int = 3, include_raw: bool = False
) -> str:
    """
    A convenience function that parses the given result from
    EXPLAIN QUERY PLAN and returns a string representation of it.
    """
    eqp = parse_explain_query_plan(result)
    raw_out = io.BytesIO()
    write_explain_query_plan(eqp, raw_out, indent=indent, include_raw=include_raw)
    return raw_out.getvalue().decode("utf-8")


def print_explain_query_plan_result(
    result: ResultItem, *, indent: int = 3, include_raw: bool = False
) -> None:
    """A convenience function that parses the given result from
    EXPLAIN QUERY PLAN and writes it to stdout.
    """
    print(
        format_explain_query_plan_result(result, indent=indent, include_raw=include_raw)
    )


from __future__ import annotations

"""
Temporal Workflow Mixins: Base
"""

import abc
import uuid
import copy
import typing as t
from ..utils import logger, null_logger
from .utils import default_id_gen_function
from lzl.types import eproperty, BaseModel, Field, field_validator

if t.TYPE_CHECKING:
    from lzl.types import AppEnv
    from lzl.logging import Logger
    from lzl.io.file import FileLike
    from ..registry import TemporalRegistry, RegistryItem
    from ..client import TemporalClient
    from temporalio.types import ParamType, ReturnType
    from lzl.ext.temporal.mixins import TemporalWorkflowT, TemporalActivityT, TemporalDispatchT, TemporalObjT


MixinKinds = t.Literal['activity', 'workflow', 'dispatch']
TemporalInputT = t.TypeVar('TemporalInputT')
TemporalReturnT = t.TypeVar('TemporalReturnT')


class IDGenOptions(BaseModel):
    """
    ID Generator Options
    """
    prefix: t.Optional[str] = None
    suffix: t.Optional[str] = None
    func: t.Optional[t.Callable[[], str]] = None
    # Field(default = default_id_gen_function, description = "The ID Generator Function")
    max_length: t.Optional[int] = 48
    id_length: t.Optional[int] = 24
    joiner: t.Optional[str] = '-'
    lower: t.Optional[bool] = True

    @field_validator('func', mode = 'before')
    def validate_func(cls, v: str | t.Callable[[], str] | None) -> t.Optional[t.Callable[[], str]]:
        """
        Validates the ID Generator Function
        """
        if v is None: return None
        if isinstance(v, str):
            from lzl.load import lazy_import
            v = lazy_import(v)
        return v

    def merge_from_config(self, config: t.Dict[str, t.Any]) -> 'IDGenOptions':
        """
        Merges the ID Generator Options from the config
        """
        new = self.model_copy(deep = True)
        if config.get('joiner'): new.joiner = config['joiner']
        if config.get('max_length'): new.max_length = config['max_length']
        if config.get('id_length'): new.id_length = config['id_length']
        if config.get('func'):
            if isinstance(config['func'], str):
                from lzl.load import lazy_import
                config['func'] = lazy_import(config['func'])
            new.func = config['func']
        
        if config.get('prefix'):
            if not new.prefix: new.prefix = config['prefix']
            elif config.get('prefix_override', False):
                new.prefix = config['prefix']
            elif config.get('prefix_append', True) and config['prefix'] not in new.prefix:
                if new.joiner not in config['prefix']: new.prefix += f'{new.joiner}{config["prefix"]}'
                else: new.prefix = config['prefix']
            if new.prefix: new.prefix.rstrip(new.joiner)
        
        if config.get('suffix'):
            if not new.suffix: new.suffix = config['suffix']
            elif config.get('suffix_override', False):
                new.suffix = config['suffix']
            elif config.get('suffix_append', True) and config['suffix'] not in new.suffix:
                if new.joiner not in config['suffix']: new.suffix += f'{new.joiner}{config["suffix"]}'
                else: new.suffix = config['suffix']
            if new.suffix: new.suffix.lstrip(new.joiner)
        return new
    

class BaseTemporalMixin(abc.ABC):
    """
    [Temporal] Base Mixin that will automatically registered
    """

    name: t.Optional[str] = None
    full_name: t.Optional[str] = None # if full_name is defined, this will override everything.

    display_name: t.Optional[str] = None
    mixin_kind: t.Optional[MixinKinds] = None

    # We don't want to use namespace as it is reserved for the Temporal namespace
    # instead we'll use workspace
    namespace: t.Optional[str] = None
    workspace: t.Optional[str] = None
    task_queue: t.Optional[str] = None

    id_gen: t.Optional[IDGenOptions] = None
    config: t.Optional[t.Dict[str, t.Union[t.Dict[str, t.Any], t.Any], t.Any]] = {}
    merge_configs: t.Optional[t.Dict[str, t.Union[t.Dict[str, t.Any], t.Any], t.Any]] = {}
    
    disable_registration: t.Optional[bool] = None
    verbosity: t.Optional[int] = None # If verbosity is < 0, logging is disabled. if verbosity is > 0, logging is enabled

    _is_subclass_: t.Optional[bool] = None
    _steps: t.Dict[str, t.Dict[str, t.Any]] = {}
    _rxtra: t.Dict[str, t.Any] = {}

    def __init_subclass__(cls, **kwargs: t.Dict[str, t.Any]):
        if cls.merge_configs:
            # This merges this subclass config with the parent config
            from lzo.utils.helpers import update_dict
            cls.config = update_dict(cls.config, cls.merge_configs)
            cls.merge_configs = {}
        
        if cls._is_subclass_: 
            cls._is_subclass_ = False
            cls = copy.deepcopy(cls)
            cls._configure_subclass_(**kwargs)
        
        elif cls.mixin_kind is not None and not cls.disable_registration and cls._should_register_mixin_(): #  and cls.name is not None: 
            cls._rxtra = {}
            cls._configure_subclass_(**kwargs)
            from lzl.ext.temporal.registry import registry
            registry.register_mixin(cls)
        return super().__init_subclass__(**kwargs)
    
    # def __new__(cls, *args, **kwargs):
    #     print(cls._rxtra)
    #     # cls._rxtra = {}
    #     return super().__new__(cls, *args, **kwargs)

    @classmethod
    def _should_register_mixin_(cls) -> bool:
        """
        Returns whether the mixin should be registered
        """
        return True
        

        
    @classmethod
    def _configure_subclass_(cls, parent_cls: t.Optional[t.Type['BaseTemporalMixin']] = None, **kwargs):
        """
        Configures the subclass
        """
        pass

    @classmethod
    def _configure_cls_(cls, **kwargs):
        """
        Configures the class
        """
        pass

    @classmethod
    def configure_registered(cls, **kwargs):
        """
        Configures the registered workflow
        """
        new = copy.deepcopy(cls)
        if kwargs.get('name'): new.name = kwargs.pop('name')
        if kwargs.get('full_name'): new.full_name = kwargs.pop('full_name')
        if 'namespace' in kwargs: new.namespace = kwargs.pop('namespace')
        if 'workspace' in kwargs: new.workspace = kwargs.pop('workspace')
        if 'task_queue' in kwargs: new.task_queue = kwargs.pop('task_queue')
        if 'mixin_kind' in kwargs: new.mixin_kind = kwargs.pop('mixin_kind')
        if 'config' in kwargs: new.config = kwargs.pop('config')
        if 'disable_registration' in kwargs: new.disable_registration = kwargs.pop('disable_registration')
        if kwargs: new._rxtra.update(kwargs)
        return new

    @classmethod
    def _configure_display_name_(cls) -> str:
        """
        Configures the Display Name for the Mixin

        Allows for more flexible customization of the display name
        """
        if cls.display_name: return cls.display_name
        if cls.full_name: 
            cls.display_name = cls.full_name
            return cls.display_name
        name_opts: t.Dict[str, bool] = cls.config.get('name_options', {})
        # logger.info(name_opts, prefix = 'Temporal')
        # logger.info(cls._rxtra, prefix = 'Temporal')
        if cls.config.get('disable_full_name'):
            if name_opts.get('use_class_name', not cls.name):
                if name_opts.get('use_full_module_name', False):
                    cls.display_name = f'{cls._rxtra["module_prefix"]}.{cls._rxtra["cls_name"]}'
                else:
                    cls.display_name = cls._rxtra['cls_name']
            else:
                cls.display_name = cls.name or cls._rxtra['base_name']
            return cls.display_name
        
        name = ''
        if name_opts.get('prefix'): name += f'{name_opts["prefix"]}.'
        if name_opts.get('include_module', True) and '__main__' not in cls._rxtra['module']: 
            if name_opts.get('use_full_module_name', False):
                name += f'{cls._rxtra["module_prefix"]}.'
            else:
                name += f'{cls._rxtra["module"]}.'
        # logger.info(f'Pre Name {name}', prefix = 'Temporal')
        if name_opts.get('include_namespace', bool(cls.namespace and not cls.workspace)) and cls.namespace: name += f'{cls.namespace}.'
        if name_opts.get('include_workspace', bool(cls.workspace)) and cls.workspace: name += f'{cls.workspace}.'
        if name_opts.get('name_extra'): name += f'{name_opts["name_extra"]}.'
        if name_opts.get('use_class_name', not cls.name):
            if cls._rxtra['module'] not in name:
                name += f'{cls._rxtra["cls_name"]}'
            else:
                name += f'{cls._rxtra["base_name"]}'
        else:
            name += f'{cls.name or cls._rxtra["base_name"]}'
        if name_opts.get('postfix'): name += f'.{name_opts["postfix"]}'
        # logger.info(f'Name {name}', prefix = 'Temporal')
        cls.display_name = name
        return cls.display_name
    
    @classmethod
    def _configure_gen_id_(cls) -> None:
        """
        Configures the ID Generator Prefix
        """
        gen_opts: t.Dict[str, t.Any] = cls.config.get('id_gen_opts', {})
        if cls.id_gen is None:
            cls.id_gen = IDGenOptions(**gen_opts)
        else:
            cls.id_gen = cls.id_gen.merge_from_config(gen_opts)
        # logger.info(f'ID Generator Options: {cls.id_gen}', prefix = 'Temporal')

    @classmethod
    def generate_id_(cls, *args, **kwargs) -> str:
        """
        Generates the ID
        """
        return default_id_gen_function(cls, *args, **kwargs)

    def generate_id(self, *args, **kwargs) -> str:
        """
        Generates the ID
        """
        if not self.id_gen: self.id_gen = IDGenOptions(**self.config.get('id_gen_opts', {}))
        return default_id_gen_function(self, *args, **kwargs)

    @classmethod
    def _update_attrs_from_config_(
        cls,
        envvar: str,
        env_mode: t.Literal['global', 'localized'] = 'localized',
        update_attrs: t.Optional[bool] = True,
        key: t.Optional[str] = None,
        app_env: t.Optional['AppEnv' | str] = None,
        **kwargs,
    ) -> t.Dict[str, t.Any] | None:
        """
        Updates the workflow attributes from the config
        """
        import os
        import yaml
        import pathlib

        if not (envfile := os.getenv(envvar)): return
        envfile = pathlib.Path(envfile)
        if not envfile.exists(): 
            logger.warning(f'Environment File does not exist: {envfile}')
            return
        envdata: t.Union[t.Dict[str, t.Any], t.Dict[t.Dict[str, t.Any]]] = yaml.safe_load(envfile.read_text())
        if cls.mixin_kind and cls.mixin_kind in envdata: envdata = envdata[cls.mixin_kind]
        if key: envdata = envdata.get(key, {})
        if env_mode == 'localized' and app_env:
            if hasattr(app_env, 'name'): envdata = envdata.get(app_env.name, {})
            else: envdata = envdata.get(app_env, {})
        if not envdata: return
        if not update_attrs: return envdata
        for key, value in envdata.items():
            if hasattr(cls, key): cls._update_attr_(key, value)

    @classmethod
    def _update_attr_(cls, key: str, value: t.Any):
        """
        Updates an attribute
        """
        attr = getattr(cls, key)
        if not attr:
            setattr(cls, key, value)
            return
        # We only care about merging dicts
        if isinstance(attr, dict):
            from lzo.utils.helpers import update_dict
            value = update_dict(attr, value)
            setattr(cls, key, value)
            return
        setattr(cls, key, value)

    
    def __init__(self, *args, **kwargs):
        """
        Initializes the workflow

        Triggers the following methods in order:

        _ _pre_init_(*args, **kwargs)
        - _init_(*args, **kwargs)
        - _configure_init_(*args, **kwargs)
        - _post_init_(*args, **kwargs)
        - _finalize_init_(*args, **kwargs)
        - _show_init_(*args, **kwargs)
        - _on_start_()
        """
        from lzo.utils import Timer
        self._extra: t.Dict[str, t.Any] = {}
        self.timer = Timer
        self.logger = logger
        self.null_logger = null_logger
        self.app_env: t.Optional['AppEnv'] = None
        self._pre_init_(*args, **kwargs)
        self._init_(*args, **kwargs)
        self._configure_init_(*args, **kwargs)
        self._post_init_(*args, **kwargs)
        self._finalize_init_(*args, **kwargs)
        self._show_init_(*args, **kwargs)
        self.registry.register_mixin(self)
        self._on_start_()

    def _on_start_(self):
        """
        Runs the on-start hook
        """
        logger.info(f'Initialized {self.name}', prefix = self.display_name, colored = True)

    def _pre_init_(self, *args, **kwargs):
        """
        Pre-Initializes the Object
        """
        pass

    def _init_(self, *args, **kwargs):
        """
        Initializes the Object
        """
        pass

    def _configure_init_(self, *args, **kwargs):
        """
        Configures the Object
        """
        pass

    def _post_init_(self, *args, **kwargs):
        """
        Post-initializes the Object
        """
        pass

    def _finalize_init_(self, *args, **kwargs):
        """
        Finalizes the Object
        """
        pass

    def _show_init_(self, *args, **kwargs):
        """
        Shows the Object
        """
        self._display_()

    
    def _display_(self, **kwargs):  # sourcery skip: use-join
        """
        Displays the configuration
        """
        pass

    @property
    def autologger(self) -> 'Logger':
        """
        Returns the auto logger
        """
        if self.verbosity:
            if self.verbosity < 0:
                return self.null_logger
            if self.verbosity > 1:
                return self.logger
        return self.logger if self.app_env and self.app_env.is_devel else self.null_logger

    @eproperty
    def registry(self) -> 'TemporalRegistry':
        """
        Returns the Temporal registry
        """
        from ..registry import registry
        return registry
    
    @property
    def registry_name(self) -> str:
        """
        Returns the registry name
        """
        return self._rxtra['registry_name']
    
    @property
    def _reg_item_(self) -> 'RegistryItem':
        """
        Returns the registry item
        """
        return self._rxtra['reg_item']

    @eproperty
    def client(self) -> t.Optional['TemporalClient']:
        """
        Returns the Temporal Client
        """
        return self._extra.get('client')
    
    

    """
    Common Class Methods
    """
    @classmethod
    def _file(cls, *paths: str) -> 'FileLike':
        """
        Returns the file
        """
        from lzl.io.file import File
        return File(*paths)

    @classmethod
    def _get_tmprl_ref_(
        cls, 
        registry_name: t.Optional[str] = None,
        module: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        workspace: t.Optional[str] = None,
        name: t.Optional[str] = None,
        display_name: t.Optional[str] = None,
        as_type: t.Optional[bool] = None,
        kind: t.Optional['MixinKinds'] = 'workflow',
    ) -> 'TemporalObjT':
        """
        Gets a Temporal Object
        """
        from ..registry import registry
        return registry.get_ref(
            registry_name = registry_name,
            module = module,
            namespace = namespace,
            name = name,
            display_name = display_name,
            as_type = as_type,
            kind = kind,
        )
        


    """
    Common Helper Methods
    """

    def get_next_cron_run(
        self,
        schedule: str,
        verbose: t.Optional[bool] = None,
        colored: t.Optional[bool] = None,
        **kwargs,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        """
        Returns the next cron run
        """
        import croniter, datetime
        utc_date = datetime.datetime.now(tz = datetime.timezone.utc)
        next_date: datetime.datetime = croniter.croniter(schedule, utc_date).get_next(datetime.datetime)
        total_seconds = (next_date - utc_date).total_seconds()
        next_interval, next_unit = total_seconds, "secs"
        # Reverse the order
        if next_interval > (60 * 60 * 24):
            next_interval /= (60 * 60 * 24)
            next_unit = "days"
        elif next_interval > (60 * 60):
            next_interval /= (60 * 60)
            next_unit = "hrs"
        elif next_interval > 60:
            next_interval /= 60
            next_unit = "mins"
        if colored: msg = f'Next Scheduled Run in |g|{next_interval:.2f} {next_unit}|e| at |g|{next_date}|e|'
        else: msg = f'Next Scheduled Run in {next_interval:.2f} {next_unit} at {next_date}'
        if verbose: self.logger.info(f'Next Scheduled Run in |g|{next_interval:.2f} {next_unit}|e| at |g|{next_date}|e| ({self.display_name})', colored = True)
        return {
            'next_date': next_date,
            'next_interval': next_interval,
            'next_unit': next_unit,
            'total_seconds': total_seconds,
            'message': msg,
        }
    
    if t.TYPE_CHECKING:
        async def run(self, arg: ParamType, *args, **kwargs) -> ReturnType:
            """
            Runs the Temporal Object
            """
            ...




from __future__ import annotations

import os
from typing import Union, Dict, Any, List, TYPE_CHECKING
from functools import lru_cache
from .utils import parse_memory_metric_to_bs, safe_float

if TYPE_CHECKING:
    from pydantic.types import ByteSize

@lru_cache()
def get_nvidia_smi_cmd() -> str:
    """
    Returns the nvidia-smi command
    """
    import platform
    from distutils import spawn
    if platform.system() == "Windows":
        # If the platform is Windows and nvidia-smi 
        # could not be found from the environment path, 
        # try to find it from system drive with default installation path
        nvidia_smi = spawn.find_executable('nvidia-smi')
        if nvidia_smi is None:
            nvidia_smi = "%s\\Program Files\\NVIDIA Corporation\\NVSMI\\nvidia-smi.exe" % os.environ['systemdrive']
    else:
        nvidia_smi = "nvidia-smi"
    return nvidia_smi


_nvidia_smi_dict_mapping = {
    'index': str,
    'name': str,
    'uuid': str,
    'utilization_gpu': safe_float,
    'utilization_memory': safe_float,
    'memory_total': parse_memory_metric_to_bs,
    'memory_used': parse_memory_metric_to_bs,
    'memory_free': parse_memory_metric_to_bs,
}

GPUData = Union[List[Dict[str, Any]], Dict[str, Union[str, int, float, 'ByteSize']]]

def map_nvidia_smi_result_values(data: List[str]) -> Dict[str, Any]:
    """
    Maps the result values to a dictionary
    """
    result: Dict[str, Any] = {
        key: _nvidia_smi_dict_mapping[key](data[n])
        for n, key in enumerate(_nvidia_smi_dict_mapping)
    }
    return result


async def aget_gpu_data() -> GPUData:
    """
    Returns the GPU Data
    """
    import asyncio
    nvidia_smi = get_nvidia_smi_cmd()
    if nvidia_smi is None: return []
    try:
        p = await asyncio.subprocess.create_subprocess_shell(
            f"{nvidia_smi} --query-gpu=index,name,uuid,utilization.gpu,utilization.memory,memory.total,memory.used,memory.free --format=csv,noheader,nounits", 
            stdout = asyncio.subprocess.PIPE, stderr = asyncio.subprocess.PIPE
        )
        stdout, _ = await p.communicate()
    except Exception:
        return []
    output = stdout.decode('UTF-8')
    if 'has failed' in output or 'command not found' in output: return []
    lines = output.split(os.linesep)
    num_devices = len(lines)-1
    gpu_data = []
    for g in range(num_devices):
        line = lines[g]
        values = line.split(', ')
        gpu_data.append(map_nvidia_smi_result_values(values))
    if len(gpu_data) == 1: gpu_data = gpu_data[0]
    return gpu_data

def get_gpu_data() -> GPUData:
    """
    Gets the GPU data
    """
    nvidia_smi = get_nvidia_smi_cmd()
    if nvidia_smi is None: return []
    import subprocess
    try:
        p = subprocess.Popen(
            f"{nvidia_smi} --query-gpu=index,name,uuid,utilization.gpu,utilization.memory,memory.total,memory.used,memory.free --format=csv,noheader,nounits", 
            stdout = subprocess.PIPE, stderr = subprocess.PIPE,
            shell = True,
        )
        stdout, _ = p.communicate()
    except Exception:
        return []
    output = stdout.decode('UTF-8')
    if 'has failed' in output or 'command not found' in output: return []
    lines = output.split(os.linesep)
    num_devices = len(lines)-1
    gpu_data = []
    for g in range(num_devices):
        line = lines[g]
        values = line.split(', ')
        gpu_data.append(map_nvidia_smi_result_values(values))
    if len(gpu_data) == 1: gpu_data = gpu_data[0]
    return gpu_data

    


from __future__ import annotations

"""
Caching Helpers
"""

import asyncio
import functools
import datetime
import typing
import contextlib
from lzl import load
from .dates import create_timestamp
from .base import is_coro_func
from typing import Callable, Dict, Any, Tuple, TypeVar, Generic, Iterable, List, Generator

if load.TYPE_CHECKING:
    import frozendict
    import async_lru
else:
    frozendict = load.LazyLoad("frozendict", install_missing=True)
    async_lru = load.LazyLoad("async_lru", install_missing=True)

def recursive_freeze(value):
    """
    Recursively freezes the object
    """
    if not isinstance(value, dict):
        return value
    for k,v in value.items():
        value[k] = recursive_freeze(v)
    return frozendict.frozendict(value)

# To unfreeze
def recursive_unfreeze(value):
    """
    Recursively unfreezes the object
    """
    if isinstance(value, frozendict.frozendict):
        value = dict(value)
        for k,v in value.items():
            value[k] = recursive_unfreeze(v)
    
    return value

def freeze_args_and_kwargs(*args, **kwargs):
    """
    Freezes the args and kwargs
    """
    args = tuple(
        recursive_freeze(arg) if isinstance(arg, dict) else arg
        for arg in args
    )
    kwargs = {k: recursive_freeze(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
    return args, kwargs


def timed_cache(
    secs: typing.Optional[int] = 60 * 60, 
    maxsize: int = 1024,
    invalidate_cache_key: typing.Optional[str] = '_invalidate_cache',
    cache_if_result: typing.Optional[typing.Any] = None,
    cache_if_type: typing.Optional[typing.Type] = None,
    exclude_none: typing.Optional[bool] = False,
    **_kwargs,
):
    """
    Wrapper for creating a expiring cached function
    args:
        secs: number of seconds to cache the function
        maxsize: maxsize of the cache
        invalidate_cache_key: key to invalidate the cache
    """
    if 'ttl' in _kwargs: secs = _kwargs.pop('ttl')
    if secs is None: secs = 60
    def wrapper_cache(func):
        if is_coro_func(func):

            @async_lru.alru_cache(maxsize = maxsize, ttl = secs)
            async def _wrapped(*args, **kwargs):
                return await func(*args, **kwargs)
            
            @functools.wraps(func)
            async def wrapped_func(*args, **kwargs):
                _invalidate = kwargs.pop(invalidate_cache_key, None) if invalidate_cache_key else None
                args, kwargs = freeze_args_and_kwargs(*args, **kwargs)
                if _invalidate is True: _wrapped.cache_invalidate(*args, **kwargs)
                result = await _wrapped(*args, **kwargs)
                if exclude_none is True and result is None:
                    _wrapped.cache_invalidate(*args, **kwargs)
                elif cache_if_result is not None and result != cache_if_result:
                    _wrapped.cache_invalidate(*args, **kwargs)
                elif cache_if_type is not None and not isinstance(result, cache_if_type):
                    _wrapped.cache_invalidate(*args, **kwargs)
                    # print(f'Invalidating cache for {func.__name__}')
                return result
            
            return wrapped_func
        
        else:
            func = functools.lru_cache(maxsize=maxsize)(func)
            func.lifetime = datetime.timedelta(seconds=secs)
            func.expiration = create_timestamp() + func.lifetime
            def _check_cache(func, invalidate: typing.Optional[bool] = None):
                if invalidate is True or create_timestamp() >= func.expiration:
                    func.cache_clear()
                    func.expiration = create_timestamp() + func.lifetime
            
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                _check_cache(func, invalidate = kwargs.pop(invalidate_cache_key, None) if invalidate_cache_key else None)
                args, kwargs = freeze_args_and_kwargs(*args, **kwargs)
                result = func(*args, **kwargs)
                if exclude_none is True and result is None:
                    func.cache_clear()
                elif cache_if_result is not None and result != cache_if_result:
                    func.cache_clear()
                elif cache_if_type is not None and not isinstance(result, cache_if_type):
                    func.cache_clear()
                return result
            
            return wrapped_func

    return wrapper_cache


    

@timed_cache(secs = 60, cache_if_result = True)
def is_in_async_loop(*args, **kwargs) -> bool:
    """
    Returns whether the app is in an async loop
    """
    with contextlib.suppress(RuntimeError):
        asyncio.get_running_loop()
        return True
    return False


import os
from lzl.logging import get_logger, change_logger_level, null_logger
# from lazyops.utils.logs import get_logger, change_logger_level, null_logger

# to prevent recursive imports, we'll just use os environ here
logger_level: str = os.getenv('LOGGER_LEVEL', 'INFO').upper()
logger = get_logger(logger_level)
logger.set_module_name('lzl.io.file', 'file', is_relative = True)



from __future__ import annotations

"""
Lazy Docker Builder Helper
"""

import os

os.environ['APP_ENV'] = 'cicd'
os.environ['DEBIAN_FRONTEND'] = 'noninteractive'

import yaml
import typer
import contextlib
import subprocess
import concurrent.futures
from pathlib import Path
from pydantic import BaseModel, model_validator, Field, PrivateAttr
from lzl.proxied import ProxyObject
from .envvars import (
    GITHUB_TOKEN,
    ASSETS_PATH,
    APP_PATH,
    REQUIREMENTS_PATH,
    PKGS_PATH,
    TEMP_PATH,
)
from .utils import (
    parse_text_file,
    run_cmd,
    add_to_env,
    build_aliases,
    echo,
)
from .static import COLOR
from typing import List, Optional, Dict, Any

BUILDS_ENABLED_REFS: List[str] = os.environ.get('BUILDS_ENABLED_REFS', 'server').split(',')
ENABLE_DEFAULT_INSTALLERS: bool = os.environ.get('ENABLE_DEFAULT_INSTALLERS', 'true').lower() in {'true', '1', 't', 'y', 'yes'}
BUILD_CONFIG_PATH = Path(os.getenv('BUILD_CONFIG_PATH', f'{TEMP_PATH}/build_config.yaml'))
DEFAULT_CONFIG_PATH = ASSETS_PATH.joinpath('default_builder.yaml')
STATE_PATH = Path(os.getenv('STATEFUL_PATH', f'{TEMP_PATH}/builder_state.yaml'))

class CustomInstaller(BaseModel):
    """
    Custom Installer

    These run during the stage 1 of the builder
    """
    name: str = Field(None, description = "The name of the custom installer")
    cmds: List[str] = Field(default_factory = list, description = "The commands to run")

    @classmethod
    def load_defaults(cls) -> List[CustomInstaller]:
        """
        Loads the default installers
        """
        default_file = ASSETS_PATH.joinpath('default_installers.yaml')
        if not default_file.exists(): return []
        data = yaml.safe_load(default_file.read_text())
        return [CustomInstaller.model_validate(item) for item in data]

    def run(self):
        """
        Runs the custom installer
        """
        echo(f'Running Custom Installer: {COLOR.BLUE}{self.name}{COLOR.END}')
        for cmdstr in self.cmds:
            os.system(cmdstr)


class CustomCommand(BaseModel):
    """
    Custom Command

    These run during the stage 2 of the builder
    """
    name: str = Field(None, description = "The name of the custom command")
    cmds: List[str] = Field(default_factory = list, description = "The commands to run")

    def run(self):
        """
        Runs the custom command
        """
        echo(f'Running Custom Command: {COLOR.BLUE}{self.name}{COLOR.END}')
        for cmdstr in self.cmds:
            os.system(cmdstr)


class BuildRef(BaseModel):
    """
    Build Ref
    """
    name: str = Field('server', description = "The name of the build reference")
    refs: List[str] = Field(default_factory = list, description = "The names of the build reference")
    enabled: Optional[bool] = Field(None, description = "Whether the build reference is enabled")
    custom_install: List[str] = Field(default_factory = list, description = "Custom Install Commands")
    custom_commands: List[str] = Field(default_factory = list, description = "Custom Commands")

class BuilderConfig(BaseModel):
    """
    Builder Config
    """
    app_name: Optional[str] = Field(os.getenv('APP_NAME'), description="The name of the app")
    refs: List[BuildRef] = Field(default_factory = list, description = "The build references")
    custom_installers: List[CustomInstaller] = Field(default_factory = list, description = "Custom Installers")
    custom_commands: List[CustomCommand] = Field(default_factory = list, description = "Custom Commands")
    enabled_refs: List[str] = Field(BUILDS_ENABLED_REFS, description = "The enabled build references")

    extra: Dict[str, Any] = Field(default_factory = dict)

    @model_validator(mode = 'after')
    def validate_refs(self):
        """
        Validate the build references
        """
        if self.extra.get('validated'): return
        for ref in self.refs:
            if ref.enabled is None:
                ref.enabled = ref.name in self.enabled_refs
        
        if ENABLE_DEFAULT_INSTALLERS:
            existing = [c.name for c in self.custom_installers]
            default_installers = CustomInstaller.load_defaults()
            for installer in default_installers:
                if installer.name not in existing:
                    self.custom_installers.append(installer)
        
        self.extra['validated'] = True
        return self

    @property
    def builds(self) -> Dict[str, BuildRef]:
        """
        Returns the build references
        """
        return {
            ref.name: ref
            for ref in self.refs if ref.enabled
        }
    
    @property
    def enabled_builds(self) -> List[str]:
        """
        Returns the enabled build services
        """
        return [ref.name for ref in self.refs if ref.enabled]

    @property
    def installers(self) -> Dict[str, CustomInstaller]:
        """
        Returns the custom installers
        """
        return {
            installer.name: installer
            for installer in self.custom_installers
        }
    
    @property
    def commands(self) -> Dict[str, CustomCommand]:
        """
        Returns the custom commands
        """
        return {
            command.name: command
            for command in self.custom_commands
        }
    
    @property
    def stage(self) -> int:
        """
        Returns the stage
        """
        return self.extra.get('stage', 0)
    
    @stage.setter
    def stage(self, value: int):
        """
        Sets the stage
        """
        self.extra['stage'] = value
        self.save()

    def show_env(self, step: str):
        """
        Show the environment
        """
        echo(f"Starting Step: {COLOR.GREEN}{step}{COLOR.END}\n")
        echo(f"[Enabled Builds]: {COLOR.BLUE}{self.enabled_builds}{COLOR.END}")

    @classmethod
    def load(cls, path: Optional[Path] = None) -> 'BuilderConfig':
        """
        Load the build config
        """
        if path is None:
            path = BUILD_CONFIG_PATH if BUILD_CONFIG_PATH.exists() else DEFAULT_CONFIG_PATH
        return cls.model_validate(yaml.safe_load(path.read_text()))
    
    def save(self):
        """
        Save the build config
        """
        BUILD_CONFIG_PATH.write_text(yaml.dump(self.model_dump(), default_flow_style = False))

    def update_config(self, path: Optional[Path] = None, **config: Any):
        """
        Update the build config
        """
        data = self.model_dump()
        if path: 
            update_data = yaml.safe_load(path.read_text())
            data.update(update_data)
        if config: data.update(config)
        BUILD_CONFIG_PATH.write_text(yaml.dump(data, default_flow_style = False))
    

    def get_apt_packages(self, ref: str) -> List[str]:
        """
        Helper for getting the apt packages for a given ref name
        """
        pkg_dir = PKGS_PATH
        if ref not in self.builds: return []
        for alias in self.builds[ref].refs:
            pkg_file = pkg_dir.joinpath(f'{alias}.txt')
            if pkg_file.exists():
                return parse_text_file(pkg_file)
        return []


    def get_pip_requirements(self, ref: str) -> Optional[str]:
        """
        Helper for getting the pip requirements file for a given ref name
        """
        pkg_dir = REQUIREMENTS_PATH.joinpath(ref)
        if ref not in self.builds: return None
        for alias in self.builds[ref].refs:
            req_file = pkg_dir.joinpath(f'{alias}.txt')
            if req_file.exists():
                req_file.write_text(req_file.read_text().replace('GITHUB_TOKEN', GITHUB_TOKEN))
                return req_file.as_posix()
            req_file = pkg_dir.joinpath(f'requirements.{alias}.txt')
            if req_file.exists():
                return req_file.as_posix()
        return None



    """
    Step 1: Install Apt Packages
    """

    @property
    def apt_pkgs(self) -> List[str]:
        """
        Returns the apt packages
        """
        if 'apt_pkgs' not in self.extra:
            self.extra['apt_pkgs'] = []
        return self.extra['apt_pkgs']


    def add_to_apt_pkgs(self, *pkg: str):
        """
        Helper for adding to the apt packages
        """
        self.apt_pkgs.extend(pkg)

    def run_apt_install(self):
        """
        Helper for running apt install
        """
        if not self.apt_pkgs: return
        _pkgs = ' '.join(list(set(self.apt_pkgs)))
        echo(f"Installing Apt Packages: {COLOR.BLUE}{_pkgs}{COLOR.END}")
        os.system(f"apt-get update && apt-get -yq install --no-install-recommends {_pkgs}")

    def get_apt_pkg_requirements(self):
        """
        Helper for getting the apt package requirements
        """
        for ref in self.enabled_builds:
            if service_pkgs := self.get_apt_packages(ref):
                echo(f'{COLOR.BLUE}[{ref}]{COLOR.END} Adding {COLOR.BOLD}{ref}{COLOR.END} requirements\n\n - {COLOR.BOLD}{service_pkgs}{COLOR.END}\n')
                self.add_to_apt_pkgs(*service_pkgs)
        self.save()


from __future__ import annotations

"""Context managers for tracking system and GPU usage during workloads."""

from .workerctx import WorkerContext
from .mlctx import MLContext

__all__ = ["WorkerContext", "MLContext"]



from __future__ import annotations

"""
Postgres Backends
"""

import typing as t
from lzl import load
from lzl.types import eproperty
from lzl.logging import logger
from ..base.backend import BaseDatabaseBackend, BackendType

if load.TYPE_CHECKING:
    import psqlpy
    from psqlpy import ConnectionPool
else:
    psqlpy = load.LazyLoad("psqlpy", install_missing=True)


class PostgresBackend(BaseDatabaseBackend):
    """
    Postgres Backend
    """

    """
    [psqlpy] Connection Pool Methods
    """

    if t.TYPE_CHECKING:
        from .config import PostgresConfig
        config: PostgresConfig

    def get_pgpool(
        self,
        readonly: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        **kwargs,
    ) -> 'ConnectionPool':
        """
        Creates a New Connection Pool
        """
        dsn = self.config.adapterless_url
        if readonly and self.config.readonly_url:
            dsn = self.config.adapterless_ro_url
        elif superuser and self.config.superuser_url:
            dsn = self.config.adapterless_superuser_url
        config = self.config.connpool.model_dump(exclude_none = True)
        config['dsn'] = dsn
        if kwargs: config.update(kwargs)
        return psqlpy.ConnectionPool(**config)
    
    @eproperty
    def pgpool(self) -> 'ConnectionPool':
        """
        Returns the Connection Pool
        """
        return self.get_pgpool()
    
    @eproperty
    def pgpool_ro(self) -> 'ConnectionPool':
        """
        [Readonly] Returns the Connection Pool
        """
        return self.get_pgpool(readonly = True)
    
    @eproperty
    def pgpool_su(self) -> 'ConnectionPool':
        """
        [Superuser] Returns the Connection Pool
        """
        return self.get_pgpool(superuser = True)


class SQLAlchemyBackend(PostgresBackend):
    """
    SQLAlchemy Backend
    """
    backend: t.Optional[BackendType] = 'sqlalchemy'


class SQLModelBackend(PostgresBackend):
    """
    SQLModel Backend
    """
    backend: t.Optional[BackendType] = 'sqlmodel'

PostgresBackendClasses = t.Union[SQLAlchemyBackend, SQLModelBackend]
PostgresBackendMap = {
    'sqlalchemy': SQLAlchemyBackend,
    'sqlmodel': SQLModelBackend,
}




from __future__ import annotations

"""
Hashing Helpers
"""

import hashlib
from lzl import load
from pydantic import BaseModel
from typing import Optional, Union, Any, TypeVar, Dict, List, TYPE_CHECKING

if load.TYPE_CHECKING:
    import xxhash
else:
    xxhash = load.LazyLoad("xxhash", install_missing = True)


PyModelType = TypeVar("PyModelType", bound=BaseModel)
ObjT = Union['PyModelType', Dict, Any]


def create_object_hash(
    obj: ObjT,
    _sep: Optional[str] = ':',
    _hash_length: Optional[int] = None,
) -> str:
    """
    Creates a hash for the object

    obj:
        the object to hash
    _sep (str, optional):
        the separator to use for the hash
        Defaults to ':'.
    _hash_length (int, optional):
        the length of the hash
        Defaults to None.
    """
    if _hash_length is None or _hash_length == 32: _hasher = xxhash.xxh3_128_hexdigest
    elif _hash_length == 16: _hasher = xxhash.xxh3_64_hexdigest
    elif _hash_length == 8: _hasher = xxhash.xxh32_hexdigest

    if isinstance(obj, dict):
        return _hasher(str(obj))
    if isinstance(obj, (list, tuple, set)):
        return f'{_sep}'.join(create_object_hash(item, _sep = _sep) for item in obj)
    if isinstance(obj, BaseModel) or hasattr(obj, "model_dump"):
        return _hasher(obj.model_dump_json(exclude_none=True))
    return _hasher(str(obj))

def create_hash_from_args_and_kwargs(
    *args,
    _typed: Optional[bool] = False,
    _key_base: Optional[tuple] = None,
    _exclude: Optional[List[str]] = None,
    _exclude_none: Optional[bool] = True,
    _sep: Optional[str] = ':',
    _hash_length: Optional[int] = None,
    **kwargs,
) -> str:
    """
    Creates a hash from the args and kwargs

    args:
        the args to hash
    _typed (bool, optional):
        if True, it will hash the args and kwargs as a typed hash
        if False, it will hash the args and kwargs as a non-typed hash
        Defaults to False.
    _key_base (tuple, optional):
        the base key to use for the hash
        Defaults to None.
    _exclude (List[str], optional):
        the keys to exclude from the hash
        Defaults to None.
    _exclude_none (bool, optional):
        if True, it will exclude any None values from the hash
        Defaults to True.
    _sep (str, optional):
        the separator to use for the hash
        Defaults to ':'.
    _hash_length (int, optional):
        the length of the hash
        Defaults to None.
    """
    hash_key = _key_base or ()
    if args: 
        hash_key += tuple(type(arg) for arg in args) if _typed else args
    if kwargs:
        if _exclude: kwargs = {k: v for k, v in kwargs.items() if k not in _exclude}
        if _exclude_none: kwargs = {k: v for k, v in kwargs.items() if v is not None}
        sorted_items = sorted(kwargs.items())
        for item in sorted_items:
            hash_key += item
    
    key = f'{_sep}'.join(str(k) for k in hash_key)
    return create_object_hash(key, _sep = _sep, _hash_length = _hash_length)





from __future__ import annotations

import abc
import typing as t
from io import BytesIO, BufferedReader # type: ignore
from lzl import load
from .types import ExtractorMode, ExtractorResult, PDFOCRStrategy, ExtractousResult

if load.TYPE_CHECKING:
    import extractous
    from lzl.io import FileLike
else:
    extractous = load.lazy_load('extractous', install_missing=True)


# https://github.com/yobix-ai/extractous/blob/336857d55920c81026936252270e46ec87ba0969/extractous-core/src/config.rs#L18
# 500_000

class Extractor(abc.ABC):
    

    @t.overload
    def __init__(
        self,
        
        # OCR Config
        density: t.Optional[int] = ...,
        depth: t.Optional[int] = ...,
        timeout_seconds: t.Optional[int] = ...,
        enable_image_preprocessing: t.Optional[bool] = ...,
        apply_rotation: t.Optional[bool] = ...,
        language: t.Optional[str] = ...,
        **kwargs,
    ):
        """
        Initializes the Extractor
            
            OCR Config Args:
                density (t.Optional[int], optional): The density. Defaults to 300.
                depth (t.Optional[int], optional): The depth. Defaults to 4.
                timeout_seconds (t.Optional[int], optional): The timeout seconds. Defaults to 130.
                enable_image_preprocessing (t.Optional[bool], optional): The enable image preprocessing. Defaults to False.
                apply_rotation (t.Optional[bool], optional): The apply rotation. Defaults to False.
                language (t.Optional[str], optional): The language. Defaults to eng.
        """
        ...

    @t.overload
    def __init__(
        self,
        
        # PDF Config
        ocr_strategy: t.Optional[PDFOCRStrategy] = ...,
        extract_inline_images: t.Optional[bool] = ...,
        extract_unique_inline_images_only: t.Optional[bool] = ...,
        extract_marked_content: t.Optional[bool] = ...,
        extract_annotation_text: t.Optional[bool] = ...,        
        **kwargs,
    ):
        """
        Initializes the Extractor
            
            PDF Config Args:
                ocr_strategy (t.Optional[PDFOCRStrategy], optional): The OCR strategy. Defaults to 'AUTO'.
                extract_inline_images (t.Optional[bool], optional): The extract inline images. Defaults to False.
                extract_unique_inline_images_only (t.Optional[bool], optional): The extract unique inline images only. Defaults to False.
                extract_marked_content (t.Optional[bool], optional): The extract marked content. Defaults to False.
                extract_annotation_text (t.Optional[bool], optional): The extract annotation text. Defaults to False.
        """
        ...
            
    @t.overload
    def __init__(
        self,
        
        # Office Config
        extract_macros: t.Optional[bool] = ...,
        include_deleted_content: t.Optional[bool] = ...,
        include_move_from_content: t.Optional[bool] = ...,
        include_shape_based_content: t.Optional[bool] = ...,
        include_headers_and_footers: t.Optional[bool] = ...,
        include_missing_rows: t.Optional[bool] = ...,
        include_slide_notes: t.Optional[bool] = ...,
        include_slide_master_content: t.Optional[bool] = ...,
        concatenate_phonetic_runs: t.Optional[bool] = ...,
        extract_all_alternatives_from_msg: t.Optional[bool] = ...,
        **kwargs,
    ):
        """
        Initializes the Extractor
            
            Office Config Args:
                extract_macros (t.Optional[bool], optional): The extract macros. Defaults to None.
                include_deleted_content (t.Optional[bool], optional): The include deleted content. Defaults to False.
                include_move_from_content (t.Optional[bool], optional): The include move from content. Defaults to False.
                include_shape_based_content (t.Optional[bool], optional): The include shape based content. Defaults to True.
                include_headers_and_footers (t.Optional[bool], optional): The include headers and footers. Defaults to False.
                include_missing_rows (t.Optional[bool], optional): The include missing rows. Defaults to False.
                include_slide_notes (t.Optional[bool], optional): The include slide notes. Defaults to True.
                include_slide_master_content (t.Optional[bool], optional): The include slide master content. Defaults to True.
                concatenate_phonetic_runs (t.Optional[bool], optional): The concatenate phonetic runs. Defaults to True.
                extract_all_alternatives_from_msg (t.Optional[bool], optional): The extract all alternatives from msg. Defaults to False.
        """
        ...

    
    @t.overload
    def __init__(
        self,
        max_string_length: t.Optional[int] = ...,
        default_encoding: t.Optional[str] = 'utf-8',
        default_chunk_size: t.Optional[int] = ...,
        default_mode: t.Optional[ExtractorMode] = 'object',
        **kwargs,
    ):
        """
        Initializes the Extractor
            
            Class Config Args:
                max_string_length (t.Optional[int], optional): The maximum string length. Defaults to 500_000.
                default_encoding (t.Optional[str], optional): The default encoding. Defaults to 'utf-8'.
                default_chunk_size (t.Optional[int], optional): The default chunk size. Defaults to 4096.
                default_mode (t.Optional[ExtractorMode], optional): The default mode. Defaults to 'object'.
        """
        ...
    

    def __init__(
        self,
        max_string_length: t.Optional[int] = None,
        default_encoding: t.Optional[str] = 'utf-8',
        default_chunk_size: t.Optional[int] = 4096,
        default_mode: t.Optional[ExtractorMode] = 'object',
        **kwargs,
    ):
        self.default_encoding = default_encoding
        self.default_chunk_size = default_chunk_size
        self.default_mode = default_mode
        self.ocr_config = extractous.TesseractOcrConfig()
        self.pdf_config = extractous.PdfParserConfig()
        self.office_config = extractous.OfficeParserConfig()
        self.set_ocr_config(**kwargs)
        self.set_pdf_config(**kwargs)
        self.set_office_config(**kwargs)
        self.api = extractous.Extractor()
        self.api.set_ocr_config(self.ocr_config)
        self.api.set_pdf_config(self.pdf_config)
        self.api.set_office_config(self.office_config)        
        if max_string_length is not None: self.api.set_extract_string_max_length(max_string_length)
    

    """
    Utilities
    """

    def _get_string_from_reader(self, reader: BufferedReader, encoding: t.Optional[str] = None, chunk_size: t.Optional[int] = None) -> str:
        """
        Gets a string from a reader
        """
        if encoding is None: encoding = self.default_encoding
        if chunk_size is None: chunk_size = self.default_chunk_size
        result = ''
        buffer = reader.read(chunk_size)
        while len(buffer) > 0:
            result += buffer.decode(encoding, errors = 'ignore')
            buffer = reader.read(chunk_size)
        return result

    def _get_output(self, reader: BufferedReader, metadata: t.Dict[str, t.Union[t.List[str], t.Any]], mode: t.Optional[ExtractorMode] = None) -> ExtractorResult:
        # sourcery skip: class-extract-method
        """
        Gets the output

        Args:
            reader (BufferedReader): The reader to get the output from
            metadata (t.Dict[str, t.Union[t.List[str], t.Any]]): The metadata to get the output from
            mode (t.Optional[ExtractorMode], optional): The mode to use. Defaults to None.

            Returns:
                t.Tuple[BufferedReader | t.Dict[str, t.Union[t.List[str], t.Any]]] |  t.Dict[str, t.Union[t.Dict[str, t.Any], t.List[str], t.Any]] | ExtractousResult: The output
        """
        if mode is None: mode = self.default_mode
        if mode == 'raw': return reader, metadata
        result = self._get_string_from_reader(reader)
        if mode == 'string': return result
        if mode == 'dict': return {'result': result, 'metadata': metadata}
        return ExtractousResult(result=result, metadata=metadata)


    """
    Mirror the Extractor API
    """

    def set_extract_string_max_length(self, max_length: int):
        """
        Sets the maximum length of the extracted string
        """
        self.api.set_extract_string_max_length(max_length)

    def extract_bytes(self, buffer: BytesIO | bytearray, mode: t.Optional[ExtractorMode] = None) -> ExtractorResult:
        """
        Extracts bytes to string
        
        Args:
            buffer (BytesIO | bytearray): The buffer to extract
            mode (t.Optional[ExtractorMode], optional): The mode to use. Defaults to None.
        
            Returns:
                t.Tuple[BufferedReader | t.Dict[str, t.Union[t.List[str], t.Any]]] | ExtractousResult: The extracted result
        """
        if isinstance(buffer, BytesIO):
            buffer = buffer.getvalue()
        reader, metadata = self.api.extract_bytes(buffer)
        return self._get_output(reader = reader, metadata = metadata, mode = mode)

    def extract_file(self, file_path: str | 'FileLike', mode: t.Optional[ExtractorMode] = None) -> ExtractorResult:
        """
        Extracts a file

        Args:
            file_path (str | FileLike): The file path or file-like object to extract
            mode (t.Optional[ExtractorMode], optional): The mode to use. Defaults to None.

        Returns:
            t.Tuple[BufferedReader | t.Dict[str, t.Union[t.List[str], t.Any]]] | ExtractousResult: The extracted result
        """
        # We're going to use the File class to get the file path
        
        from lzl.io import File
        file_path = File(file_path)
        buffer = bytearray(file_path.read())
        return self.extract_bytes(buffer, mode = mode)

    def extract_file_to_string(self, file_path: str | 'FileLike', mode: t.Optional[ExtractorMode] = None) -> ExtractorResult:
        """
        Extracts a File to string

        Args:
            file_path (str | FileLike): The file path or file-like object to extract
            mode (t.Optional[ExtractorMode], optional): The mode to use. Defaults to None.

        Returns:
            t.Tuple[BufferedReader | t.Dict[str, t.Union[t.List[str], t.Any]]] | ExtractousResult: The extracted result
        """
        reader, metadata = self.api.extract_file_to_string(file_path)
        return self._get_output(reader = reader, metadata = metadata, mode = mode)


    def extract_url(self, url: str, mode: t.Optional[ExtractorMode] = None) -> ExtractorResult:
        """
        Extracts a URL

        Args:
            url (str): The URL to extract
            mode (t.Optional[ExtractorMode], optional): The mode to use. Defaults to None.

        Returns:
            t.Tuple[BufferedReader | t.Dict[str, t.Union[t.List[str], t.Any]]] | ExtractousResult: The extracted result
        """
        reader, metadata = self.api.extract_url(url)
        metadata['url'] = url
        return self._get_output(reader = reader, metadata = metadata, mode = mode)

    def extract_url_to_string(self, url: str, mode: t.Optional[ExtractorMode] = None) -> ExtractorResult:
        """
        Extracts a URL to string

        Args:
            url (str): The URL to extract
            mode (t.Optional[ExtractorMode], optional): The mode to use. Defaults to None.

        Returns:
            t.Tuple[BufferedReader | t.Dict[str, t.Union[t.List[str], t.Any]]] | ExtractousResult: The extracted result
        """
        reader, metadata = self.api.extract_url_to_string(url)
        metadata['url'] = url
        return self._get_output(reader = reader, metadata = metadata, mode = mode)


    """
    Configuration Methods
    """
            
    def set_ocr_config(
        self,
        density: t.Optional[int] = None,
        depth: t.Optional[int] = None,
        timeout_seconds: t.Optional[int] = None,
        enable_image_preprocessing: t.Optional[bool] = None,
        apply_rotation: t.Optional[bool] = None,
        language: t.Optional[str] = None,
        **kwargs,
    ):
        """
        Sets the OCR config

        Args:
            density (t.Optional[int], optional): The density. Defaults to 300.
            depth (t.Optional[int], optional): The depth. Defaults to 4.
            timeout_seconds (t.Optional[int], optional): The timeout seconds. Defaults to 130.
            enable_image_preprocessing (t.Optional[bool], optional): The enable image preprocessing. Defaults to False.
            apply_rotation (t.Optional[bool], optional): The apply rotation. Defaults to False.
            language (t.Optional[str], optional): The language. Defaults to eng.
        
        """
        if density is not None: self.ocr_config.set_density(density)
        if depth is not None: self.ocr_config.set_depth(depth)
        if timeout_seconds is not None: self.ocr_config.set_timeout_seconds(timeout_seconds)
        if enable_image_preprocessing is not None: self.ocr_config.set_enable_image_preprocessing(enable_image_preprocessing)
        if apply_rotation is not None: self.ocr_config.set_apply_rotation(apply_rotation)
        if language is not None: self.ocr_config.set_language(language)

    def set_pdf_config(
        self,
        ocr_strategy: t.Optional[PDFOCRStrategy] = None,
        extract_inline_images: t.Optional[bool] = None,
        extract_unique_inline_images_only: t.Optional[bool] = None,
        extract_marked_content: t.Optional[bool] = None,
        extract_annotation_text: t.Optional[bool] = None,
        **kwargs,
    ):
        """
        Sets the PDF config

        Args:
            ocr_strategy (t.Optional[PDFOCRStrategy], optional): The OCR strategy. Defaults to 'AUTO'.
            extract_inline_images (t.Optional[bool], optional): The extract inline images. Defaults to False.
            extract_unique_inline_images_only (t.Optional[bool], optional): The extract unique inline images only. Defaults to False.
            extract_marked_content (t.Optional[bool], optional): The extract marked content. Defaults to False.
            extract_annotation_text (t.Optional[bool], optional): The extract annotation text. Defaults to False.
        """
        if ocr_strategy is not None: self.pdf_config.set_ocr_strategy(ocr_strategy)
        if extract_inline_images is not None: self.pdf_config.set_extract_inline_images(extract_inline_images)
        if extract_unique_inline_images_only is not None: self.pdf_config.set_extract_unique_inline_images_only(extract_unique_inline_images_only)
        if extract_marked_content is not None: self.pdf_config.set_extract_marked_content(extract_marked_content)
        if extract_annotation_text is not None: self.pdf_config.set_extract_annotation_text(extract_annotation_text)

    def set_office_config(
        self,
        extract_macros: t.Optional[bool] = None,
        include_deleted_content: t.Optional[bool] = None,
        include_move_from_content: t.Optional[bool] = None,
        include_shape_based_content: t.Optional[bool] = None,
        include_headers_and_footers: t.Optional[bool] = None,
        include_missing_rows: t.Optional[bool] = None,
        include_slide_notes: t.Optional[bool] = None,
        include_slide_master_content: t.Optional[bool] = None,
        concatenate_phonetic_runs: t.Optional[bool] = None,
        extract_all_alternatives_from_msg: t.Optional[bool] = None,
        **kwargs,
    ):
        """
        Sets the Office config

        Args:
            extract_macros (t.Optional[bool], optional): The extract macros. Defaults to None.
            include_deleted_content (t.Optional[bool], optional): The include deleted content. Defaults to False.
            include_move_from_content (t.Optional[bool], optional): The include move from content. Defaults to False.
            include_shape_based_content (t.Optional[bool], optional): The include shape based content. Defaults to True.
            include_headers_and_footers (t.Optional[bool], optional): The include headers and footers. Defaults to False.
            include_missing_rows (t.Optional[bool], optional): The include missing rows. Defaults to False.
            include_slide_notes (t.Optional[bool], optional): The include slide notes. Defaults to True.
            include_slide_master_content (t.Optional[bool], optional): The include slide master content. Defaults to True.
            concatenate_phonetic_runs (t.Optional[bool], optional): The concatenate phonetic runs. Defaults to True.
            extract_all_alternatives_from_msg (t.Optional[bool], optional): The extract all alternatives from msg. Defaults to False.
        """
        if extract_macros is not None: self.office_config.set_extract_macros(extract_macros)
        if include_deleted_content is not None: self.office_config.set_include_deleted_content(include_deleted_content)
        if include_move_from_content is not None: self.office_config.set_include_move_from_content(include_move_from_content)
        if include_shape_based_content is not None: self.office_config.set_include_shape_based_content(include_shape_based_content)
        if include_headers_and_footers is not None: self.office_config.set_include_headers_and_footers(include_headers_and_footers)
        if include_missing_rows is not None: self.office_config.set_include_missing_rows(include_missing_rows)
        if include_slide_notes is not None: self.office_config.set_include_slide_notes(include_slide_notes)
        if include_slide_master_content is not None: self.office_config.set_include_slide_master_content(include_slide_master_content)
        if concatenate_phonetic_runs is not None: self.office_config.set_concatenate_phonetic_runs(concatenate_phonetic_runs)
        if extract_all_alternatives_from_msg is not None: self.office_config.set_extract_all_alternatives_from_msg(extract_all_alternatives_from_msg)

    def set_encoding(
        self,
        encoding: t.Optional[str] = None,
        **kwargs,
    ):
        """
        Sets the encoding
        
        Args:
            encoding (t.Optional[str], optional): The encoding. Defaults to None.
        """
        if encoding is not None: self.api.set_encoding(encoding)

    




from __future__ import annotations


import enum
import json
import time
import asyncio
import contextlib
from lzl.api import aiohttpx
from typing import Optional, Type, Any, Union, List, Dict, Iterator, AsyncIterator, Generator, AsyncGenerator, TYPE_CHECKING

from lzo.types import (
    Literal, 
    eproperty, 
    BaseModel,
    validator, 
    root_validator, 
    field_validator,
    Field,
    PYDANTIC_VERSION, 
)
from lzl.api.openai.version import DEFAULT_COMPLETION_MODEL
from lzl.api.openai.types.handlers import ModelContextHandler
from lzl.api.openai.types.base import BaseResource, Usage
from lzl.api.openai.types.responses import BaseResponse
from lzl.api.openai.types.routes import BaseRoute
from lzl.api.openai.utils import logger, parse_stream, aparse_stream


__all__ = [
    'CompletionChoice',
    'CompletionObject',
    'CompletionResponse',
    'CompletionRoute',
]


class CompletionKind(str, enum.Enum):
    # Only have one choice
    CONTENT = 'content'

class StreamedCompletionChoice(BaseResource):
    kind: CompletionKind = CompletionKind.CONTENT
    value: str


class CompletionChoice(BaseResource):
    text: str
    index: int
    logprobs: Optional[Any] = None
    finish_reason: Optional[str] = None



class CompletionObject(BaseResource):
    prompt: Union[List[str], str] = '<|endoftext|>'
    model: Optional[str] = DEFAULT_COMPLETION_MODEL
    # prompt: Optional[Union[List[str], str]] = '<|endoftext|>'
    suffix: Optional[str] = None
    max_tokens: Optional[int] = 16
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    stream: Optional[bool] = False
    logprobs: Optional[int] = None
    echo: Optional[bool] = False
    stop: Optional[Union[List[str], str]] = None
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    best_of: Optional[int] = 1
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None

    validate_max_tokens: Optional[bool] = Field(default = False, exclude = True)
    # validate_model_aliases: Optional[bool] = Field(default = False, exclude = True)

    # @validator('model', pre=True, always=True)
    @field_validator('model')
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'):
                v = values.get('engine')
            elif values.get('deployment'):
                v = values.get('deployment')
        with contextlib.suppress(KeyError):
            v = ModelContextHandler.resolve_model_name(v)
        # if values.get('validate_model_aliases', False):
        #     v = ModelContextHandler[v].name
        return v

    
    @validator('temperature')
    def validate_temperature(cls, v: float) -> float:
        """
        Min Temperature is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-temperature
        """
        return None if v is None else max(0.0, v)
    
    @validator('top_p')
    def validate_top_p(cls, v: float) -> float:
        """
        Min Top Probability is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-top-probability
        """
        return None if v is None else max(0.0, v)

    @validator('logprobs')
    def validate_logprobs(cls, v: int) -> int:
        """
        Max logprobs is 5
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-logprobs
        """
        return None if v is None else max(0, min(v, 5))
    
    @validator('presence_penalty')
    def validate_presence_penalty(cls, v: float) -> float:
        """
        Min Presence Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-presence-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    
    @validator('frequency_penalty')
    def validate_frequency_penalty(cls, v: float) -> float:
        """
        Min Frequency Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-frequency-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, **kwargs)


    

class CompletionResponse(BaseResponse):
    choices: Optional[List[CompletionChoice]] = None
    choice_model: Optional[Type[BaseResource]] = CompletionChoice
    input_object: Optional[CompletionObject] = None

    @eproperty
    def text(self) -> str:
        """
        Returns the text for the completions
        """
        if self.choices_results:
            return ''.join([choice.text for choice in self.choices])
        return self.response.text
    
    @eproperty
    def openai_model(self):
        """
        Returns the model for the completions
        """
        return self.headers.get('openai-model', self.completion_model)

    @eproperty
    def completion_model(self):
        """
        Returns the model for the completions
        """
        return self.input_object.model or None

    def _validate_usage(self):
        """
        Validate usage
        """
        if self.usage and self.usage.total_tokens: return
        if self.response.status_code == 200:
            self.usage = Usage(
                prompt_tokens = ModelContextHandler.count_tokens(self.input_object.prompt),
                completion_tokens = ModelContextHandler.count_tokens(self.text),
            )
            self.usage.total_tokens = self.usage.prompt_tokens + self.usage.completion_tokens

    @eproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """ 
        self._validate_usage()
        return ModelContextHandler.get_consumption_cost(
            model_name = self.openai_model,
            usage = self.usage,
        )

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, **kwargs)

    def parse_stream_item(self, item: Union[Dict, Any], **kwargs) -> Optional[StreamedCompletionChoice]:
        """
        Parses a single stream item
        """
        choice = item['choices'][0]
        if choice['finish_reason'] == 'stop':
            return None
        return StreamedCompletionChoice(
            value = choice['text']
        )

    def handle_stream(
        self,
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> Iterator[Dict]:
        """
        Handles the streaming response
        """
        texts = {}
        for line in parse_stream(response):
            try:
                item = json.loads(line)
                if streaming:
                    yield item
                self.handle_stream_metadata(item)
                for n, choice in enumerate(item['choices']):
                    if not texts.get(n):
                        texts[n] = {
                            'index': choice['index'],
                            'text': choice['text'],
                        }
                        self.usage.completion_tokens += 1
                    elif choice['finish_reason'] != 'stop':
                        texts[n]['text'] += choice['text']
                        self.usage.completion_tokens += 1

                    else:
                        texts[n]['finish_reason'] = choice['finish_reason']
                        compl = texts.pop(n)
                        if streaming:
                            self.handle_resource_item(item = compl)
                        else:
                            yield compl

            except Exception as e:
                logger.error(f'Error: {line}: {e}')

        self._stream_consumed = True
        for remaining_text in texts.values():
            if streaming:
                self.handle_resource_item(item = remaining_text)
            else:
                yield remaining_text
        self.usage.total_tokens = self.usage.completion_tokens
    

    async def ahandle_stream(
        self,
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> AsyncIterator[Dict]:
        """
        Handles the streaming response
        """
        texts = {}
        async for line in aparse_stream(response):
            try:
                item = json.loads(line)
                if streaming:
                    yield item
                self.handle_stream_metadata(item)
                for n, choice in enumerate(item['choices']):
                    if not texts.get(n):
                        texts[n] = {
                            'index': choice['index'],
                            'text': choice['text'],
                        }
                        self.usage.completion_tokens += 1
                    elif choice['finish_reason'] != 'stop':
                        texts[n]['text'] += choice['text']
                        self.usage.completion_tokens += 1

                    else:
                        texts[n]['finish_reason'] = choice['finish_reason']
                        compl = texts.pop(n)
                        if streaming:
                            self.handle_resource_item(item = compl)
                        else:
                            yield compl

            except Exception as e:
                logger.error(f'Error: {line}: {e}')

        self._stream_consumed = True
        for remaining_text in texts.values():
            if streaming:
                self.handle_resource_item(item = remaining_text)
            else:
                yield remaining_text
        self.usage.total_tokens = self.usage.completion_tokens

    if TYPE_CHECKING:
        def stream(self, **kwargs) -> Generator[StreamedCompletionChoice, None, None]:
            """
            Streams the response
            """
            ...

        async def astream(self, **kwargs) -> AsyncGenerator[StreamedCompletionChoice, None]:
            """
            Streams the response
            """
            ...
    

class CompletionRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = CompletionObject
    response_model: Optional[Type[BaseResource]] = CompletionResponse

    api_resource: Optional[str] = Field(default = 'completions')
    root_name: Optional[str] = Field(default = 'completion')

    # @eproperty
    # def api_resource(self):
    #     return 'completions'

    # @eproperty
    # def root_name(self):
    #     return 'completion'
    
    def create(
        self, 
        input_object: Optional[CompletionObject] = None,
        parse_stream: Optional[bool] = True,
        timeout: Optional[int] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        header_cache_keys: Optional[List[str]] = None,
        **kwargs
    ) -> CompletionResponse:
        """
        Creates a completion for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.completions.create(
        >>>    prompt = 'say this is a test',
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `text-curie-001`
        
        :prompt (optional): The prompt(s) to generate completions for, encoded as 
        a string, array of strings, array of tokens, or array of token arrays.
        Note that `<|endoftext|>` is the document separator that the model sees during 
        training, so if a prompt is not specified the model will generate as if from 
        the beginning of a new document.
        Default: `<|endoftext|>`

        :suffix (optional): The suffix that comes after a completion of inserted text.
        Default: `None`

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :echo (optional): Echo back the prompt in addition to the completion
        Default: `False`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :best_of (optional): Generates `best_of` completions server-side and returns 
        the "best" (the one with the highest log probability per token). Results cannot be streamed.
        When used with `n`, `best_of` controls the number of candidate completions and n specifies how 
        many to return – `best_of` must be greater than `n`.
        Note: Because this parameter generates many completions, it can quickly consume your token quota. 
        Use carefully and ensure that you have reasonable settings for `max_tokens` and `stop`.
        Default: `1`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Returns: `CompletionResult`
        """
        return super().create(
            input_object = input_object,
            parse_stream = parse_stream,
            timeout = timeout,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            header_cache_keys = header_cache_keys,
            **kwargs
        )


    async def async_create(
        self, 
        input_object: Optional[CompletionObject] = None,
        parse_stream: Optional[bool] = True,
        timeout: Optional[int] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        header_cache_keys: Optional[List[str]] = None,
        **kwargs
    ) -> CompletionResponse:
        """
        Creates a completion for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.completions.async_create(
        >>>    prompt = 'say this is a test',
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `text-curie-001`
        
        :prompt (optional): The prompt(s) to generate completions for, encoded as 
        a string, array of strings, array of tokens, or array of token arrays.
        Note that `<|endoftext|>` is the document separator that the model sees during 
        training, so if a prompt is not specified the model will generate as if from 
        the beginning of a new document.
        Default: `<|endoftext|>`

        :suffix (optional): The suffix that comes after a completion of inserted text.
        Default: `None`

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :echo (optional): Echo back the prompt in addition to the completion
        Default: `False`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :best_of (optional): Generates `best_of` completions server-side and returns 
        the "best" (the one with the highest log probability per token). Results cannot be streamed.
        When used with `n`, `best_of` controls the number of candidate completions and n specifies how 
        many to return – `best_of` must be greater than `n`.
        Note: Because this parameter generates many completions, it can quickly consume your token quota. 
        Use carefully and ensure that you have reasonable settings for `max_tokens` and `stop`.
        Default: `1`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Returns: `CompletionResult`
        """
        return await super().async_create(
            input_object = input_object,
            parse_stream = parse_stream,
            timeout = timeout,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            header_cache_keys = header_cache_keys,
            **kwargs
        )
    



from __future__ import annotations

"""
FastAPI Helpers
"""

from lzl.proxied import ProxyObject
from typing import Optional, List, Dict, Any, Union, Type, Callable, TYPE_CHECKING


if TYPE_CHECKING:
    from lzl.io import JsonSerializer

serializer: 'JsonSerializer' = ProxyObject(
    obj_getter = 'lzl.io.ser.get_serializer',
    obj_args = ('json',)
)



from __future__ import annotations


import abc
import datetime
import typing as t

if t.TYPE_CHECKING:
    from lzl.types import AppEnv

def parse_db_config(
    config: t.Dict[str, t.Dict[str, t.Union[str, t.Dict[str, str]]]],
    env_name: t.Optional[t.Union[str, 'AppEnv']] = 'local',
    in_cluster: t.Optional[bool] = None,
    default_adapter: t.Optional[str] = 'postgresql+asyncpg',
    **kwargs,
) -> t.Dict[str, str]:
    """
    Constructs the Database URLs from a YAML Configuration
    
    Function to transform
    users:
        production: user:...
        development: user_development:...
        staging: user_staging:...
        local: user_local:...
        superuser: postgres:...
    database: app
    adapter: postgresql+asyncpg
    endpoints:
        public: appdb.com:5432
        cluster:
            rw: appdb-v1-cluster-rw.db.svc.cluster.local:5432
            ro: appdb-v1-cluster-ro.db.svc.cluster.local:5432
            prw: appdb-v1-cluster-pooler-rw.db.svc.cluster.local:5432
            pro: appdb-v1-cluster-pooler-ro.db.svc.cluster.local:5432

    to (Public)

    url: postgresql://user:...@appdb.com:5432/app
    superuser_url: postgresql://postgres:...@appdb.com:5432/app

    to (Cluster)
 
    url: postgresql://user:...@appdb-v1-cluster-pooler-rw.db.svc.cluster.local:5432/app
    readonly_url: postgresql://user:...@appdb-v1-cluster-pooler-ro.db.svc.cluster.local:5432/app
    superuser_url: postgresql://postgres:...@appdb-v1-cluster-rw.db.svc.cluster.local:5432/app
    """
    if hasattr(env_name, 'name'): env_name = env_name.name
    if in_cluster is None:
        from lzo.utils.system import is_in_kubernetes
        in_cluster = is_in_kubernetes()
    
    uri_base = f'{config.get("adapter", default_adapter)}://'
    db = config['database']
    results = {}
    if user := config['users'].get(env_name):
        user_uri_base = f'{uri_base}{user}'
        if in_cluster:
            cluster_eps = config['endpoints']['cluster']
            if isinstance(cluster_eps, str): 
                results['url'] = f'{user_uri_base}@{cluster_eps}/{db}'
            
            else:
                if cluster_eps.get('prw'):
                    results['url'] = f'{user_uri_base}@{cluster_eps["prw"]}/{db}'
                elif cluster_eps.get('rw'):
                    results['url'] = f'{user_uri_base}@{cluster_eps["rw"]}/{db}'
                
                if cluster_eps.get('pro'):
                    results['readonly_url'] = f'{user_uri_base}@{cluster_eps["pro"]}/{db}'
                elif cluster_eps.get('ro'):
                    results['readonly_url'] = f'{user_uri_base}@{cluster_eps["ro"]}/{db}'
        
        elif isinstance(config['endpoints']['public'], dict):
            public_eps = config['endpoints']['public']
            if public_eps.get('prw'):
                results['url'] = f'{user_uri_base}@{public_eps["prw"]}/{db}'
            elif public_eps.get('rw'):
                results['url'] = f'{user_uri_base}@{public_eps["rw"]}/{db}'
            
            if public_eps.get('pro'):
                results['readonly_url'] = f'{user_uri_base}@{public_eps["pro"]}/{db}'
            elif public_eps.get('ro'):
                results['readonly_url'] = f'{user_uri_base}@{public_eps["ro"]}/{db}'
        
        else:
            results['url'] = f'{user_uri_base}@{config["endpoints"]["public"]}/{db}'
    
    if superuser := config['users'].get('superuser'):
        superuser_uri_base = f'{uri_base}{superuser}'
        if in_cluster:
            cluster_eps = config['endpoints']['cluster']
            if isinstance(cluster_eps, str):
                results['superuser_url'] = f'{superuser_uri_base}@{cluster_eps}/{db}'
            elif cluster_eps.get('rw'):
                results['superuser_url'] = f'{superuser_uri_base}@{cluster_eps["rw"]}/{db}'
            elif cluster_eps.get('prw'):
                results['superuser_url'] = f'{superuser_uri_base}@{cluster_eps["prw"]}/{db}'
        elif isinstance(config['endpoints']['public'], dict):
            public_eps = config['endpoints']['public']
            if public_eps.get('rw'):
                results['superuser_url'] = f'{superuser_uri_base}@{public_eps["rw"]}/{db}'
            elif public_eps.get('prw'):
                results['superuser_url'] = f'{superuser_uri_base}@{public_eps["prw"]}/{db}'
        else:
            results['superuser_url'] = f'{superuser_uri_base}@{config["endpoints"]["public"]}/{db}'
    return results



def build_sql_metadata_filter(
    table: str,
    conditional: t.Optional[str] = 'AND',
    metadata_key: t.Optional[str] = '_metadata',
    include_table_name: t.Optional[bool] = False,
    **filters: t.Dict[str, t.Union[int, float, datetime.datetime, t.Dict, t.List, t.Any]]
) -> str:
    """
    Constructs the WHERE clause for the `_metadata` property because it's a `jsonb` field
    """
    q = ""
    if include_table_name and table: metadata_key = f'{table}.{metadata_key}'
    for key, value in filters.items():
        if 'date' in key:
            # Handle dates
            if isinstance(value, list):
                # Handle date ranges
                if isinstance(value[0], datetime.datetime):
                    value[0] = value[0].strftime("%Y-%m-%d %H:%M:%S")
                    value[1] = value[1].strftime("%Y-%m-%d %H:%M:%S")
                q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ BETWEEN {value[0]} AND {value[1]} || @ == null)' {conditional} "
            else:
                # Handle single dates
                op = '>' if key == "open_date" else '<='
                q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ {op} {value} || @ == null)' {conditional} "
            continue

        # Handle Ints and Floats
        if isinstance(value, (int, float)):
            op = '>=' if 'min' in key else '<='
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ {op} {value} || @ == null)' {conditional} "
            continue

        # Handle Strings
        if isinstance(value, str):
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ ILIKE %'{value}'% || @ == null)' {conditional} "
            continue

        # Handle Lists
        if isinstance(value, list):
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ && {value} || @ == null)' {conditional} "
            continue

        # Handle Dicts
        if isinstance(value, dict):
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ ?& {value.keys()} || @ == null)' {conditional} "
            continue

    q = q[:-len(conditional)-1]
    return q


class SQLAlchemyUtilities(abc.ABC):
    """
    Abstract Base Class for SQLAlchemy Utilities
    """
    def __init__(self):
        """
        Initializes the SQLAlchemy Utilities
        """
        from sqlalchemy.orm import defer
        from sqlalchemy.sql.expression import (
            text, select, update, delete,
            funcfilter, lambda_stmt, func, modifier, bindparam,
            exists, desc, asc, case, cast, literal,
            collate, distinct, extract, false, null, nulls_first, nulls_last, true,
            over, between, lateral, try_cast, alias, type_coerce, within_group,
            intersect, intersect_all, outerjoin, union, union_all, table, values,
            all_, and_, any_, not_, or_, tuple_, except_, except_all,
        )
        from sqlalchemy.dialects.postgresql import insert, array_agg
        from sqlalchemy.orm import selectinload
        self.defer = defer
        self.text = text
        self.select = select
        self.update = update
        self.delete = delete
        self.funcfilter = funcfilter
        self.lambda_stmt = lambda_stmt
        self.func = func
        self.modifier = modifier
        self.bindparam = bindparam

        self.exists = exists
        self.desc = desc
        self.asc = asc
        self.case = case
        self.cast = cast
        self.literal = literal

        self.collate = collate
        self.distinct = distinct
        self.extract = extract
        self.false = false
        self.null = null
        self.nulls_first = nulls_first
        self.nulls_last = nulls_last
        self.true = true
        self.over = over
        self.between = between
        self.lateral = lateral
        self.try_cast = try_cast
        self.alias = alias
        self.type_coerce = type_coerce
        self.within_group = within_group

        self.intersect = intersect
        self.intersect_all = intersect_all
        self.outerjoin = outerjoin
        self.union = union
        self.union_all = union_all
        self.table = table
        self.values = values

        self.and_ = and_
        self.all_ = all_
        self.any_ = any_
        self.not_ = not_
        self.or_ = or_
        self.tuple_ = tuple_
        self.except_ = except_
        self.except_all = except_all

        self.select = select
        self.update = update
        self.delete = delete
        self.insert = insert
        self.selectinload = selectinload
        self.array_agg = array_agg
        self.postinit()

    def postinit(self):
        """
        Post Initialization
        """
        pass



"""
Serialization Utilities
"""

import abc
import json
import codecs
import datetime
import dataclasses
import contextlib
from uuid import UUID
from enum import Enum
from lzl.logging import logger
from lzl.load import lazy_import
from lzl.types import BaseModel
from typing import Optional, Union, Any, Dict, List, Tuple, Callable, Type, Mapping, TypeVar, Literal, TYPE_CHECKING

try:
    import numpy as np
    np_version = np.__version__
    if np_version.startswith("2."):
        np_float_types = (np.float16, np.float32, np.float64)
    else:
        np_float_types = (np.float_, np.float16, np.float32, np.float64)
    np_int_types = (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)
    
except ImportError:
    np = None

import pickle
try:
    import cloudpickle
    _cloudpicke_available = True
except ImportError:
    _cloudpicke_available = False

try:
    import dill
    _dill_available = True
except ImportError:
    _dill_available = False

if _cloudpicke_available:
    default_pickle = cloudpickle
elif _dill_available:
    default_pickle = dill
else:
    default_pickle = pickle

SerMode = Literal['auto', 'raw']
SerializableObject = TypeVar('SerializableObject')
serialization_class_registry: Dict[str, Type[SerializableObject]] = {}

_alias_schema_mapping: Dict[str, str] = {}
_null = b"\x00"
_null2 = _null * 2
_null3 = _null * 3

def guess_json_utf(data: bytes) -> Optional[str]:
    # JSON always starts with two ASCII characters, so detection is as
    # easy as counting the nulls and from their location and count
    # determine the encoding. Also detect a BOM, if present.
    sample = data[:4]
    if sample in (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE):
        return "utf-32"  # BOM included
    if sample[:3] == codecs.BOM_UTF8:
        return "utf-8-sig"  # BOM included, MS style (discouraged)
    if sample[:2] in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
        return "utf-16"  # BOM included
    nullcount = sample.count(_null)
    if nullcount == 0:
        return "utf-8"
    if nullcount == 2:
        if sample[::2] == _null2:  # 1st and 3rd are null
            return "utf-16-be"
        if sample[1::2] == _null2:  # 2nd and 4th are null
            return "utf-16-le"
    elif nullcount == 3:
        if sample[:3] == _null3:
            return "utf-32-be"
        if sample[1:] == _null3:
            return "utf-32-le"
    return None


def is_primitive(value, exclude_bytes: Optional[bool] = False) -> bool:
    """
    Check if a value is a primitive type
    """
    if exclude_bytes and isinstance(value, bytes): return False
    return isinstance(value, (int, float, bool, str, bytes, type(None)))


def get_object_classname(obj: SerializableObject) -> str:
    """
    Get the classname of an object
    """
    return f"{obj.__class__.__module__}.{obj.__class__.__name__}"

def get_object_class(name: str) -> Type[SerializableObject]:
    """
    Get the class of an object
    """
    global serialization_class_registry
    if name not in serialization_class_registry:
        serialization_class_registry[name] = lazy_import(name)
    return serialization_class_registry[name]

def register_object_class(obj: SerializableObject) -> str:
    """
    Register the object class
    """
    global serialization_class_registry
    obj_class_name = get_object_classname(obj)
    if obj_class_name not in serialization_class_registry:
        serialization_class_registry[obj_class_name] = obj.__class__
    return obj_class_name

def register_schema_mapping(schemas: Dict[str, str]):
    """
    Register the schema mapping
    """
    global _alias_schema_mapping
    _alias_schema_mapping.update(schemas)


def parse_list_str(
    line: Optional[Union[List[str], str]],
    default: Optional[List[str]] = None,
    seperators: Optional[List[str]] = None,
) -> Optional[List[str]]:
    """
    Try to parse a string as a list of strings

    Args:
        line (typing.Optional[typing.Union[typing.List[str], str]]): [description]
        default (typing.Optional[typing.List[str]], optional): [description]. Defaults to None.
        seperators (typing.Optional[typing.List[str]], optional): [description]. Defaults to None.
    
    """
    if line is None: return default
    if seperators is None: seperators = [',', '|', ';']
    if isinstance(line, list): return line
    if isinstance(line, str):
        if '[' in line and ']' in line:
            if '"' in line or "'" in line:
                try:
                    line = json.loads(line)
                    return line
                except Exception: line = line.replace("'", '').replace('"', '')
            line = line.replace('[', '').replace(']', '')
        for seperator in seperators:
            if seperator in line:
                return line.split(seperator)
        line = [line]
    return line
    

def serialize_object(
    obj: SerializableObject,
    mode: Optional[SerMode] = 'auto',
    **kwargs
) -> Union[Dict[str, Any], List[Dict[str, Any]], Any]:
    # sourcery skip: extract-duplicate-method
    """
    Helper to serialize an object

    Args:
        obj: the object to serialize

    Returns:
        the serialized object in dict
        {
            "__type__": ...,
            "value": ...,
        }
    """
    if obj is None: return None

    if isinstance(obj, BaseModel) or hasattr(obj, 'model_dump'):
        obj_class_name = register_object_class(obj)
        obj_value = obj.model_dump(mode = 'json', round_trip = True, **kwargs)
        if mode == 'raw': return obj_value
        return {
            "__type__": "pydantic",
            "__class__": obj_class_name,
            "value": obj_value,
        }

    # Move this to the top before primitives
    if np is not None:
        # if isinstance(obj, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
        if isinstance(obj, np_int_types):
            if mode == 'raw': return int(obj)
            obj_class_name = register_object_class(obj)
            return {
                "__type__": "numpy",
                "__class__": obj_class_name,
                "value": int(obj),
            }


        # if isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
        if isinstance(obj, np_float_types):
            if mode == 'raw': return float(obj)
            obj_class_name = register_object_class(obj)
            return {
                "__type__": "numpy",
                "__class__": obj_class_name,
                "value": float(obj),
            }


    if is_primitive(obj, exclude_bytes = True):
        return obj

    if isinstance(obj, (list, tuple)):
        return [serialize_object(item, mode = mode, **kwargs) for item in obj]

    if isinstance(obj, dict):
        if "__type__" in obj: 
            return obj['value'] if mode == 'raw' and obj.get('value') else obj
        return {key: serialize_object(value, mode = mode, **kwargs) for key, value in obj.items()}

    if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
        if mode == 'raw': return obj.isoformat()
        return {
            "__type__": "datetime",
            "value": obj.isoformat(),
        }

    if isinstance(obj, datetime.timedelta):
        if mode == 'raw': return obj.total_seconds()
        return {
            "__type__": "timedelta",
            "value": obj.total_seconds(),
        }

    if isinstance(obj, dataclasses.InitVar) or dataclasses.is_dataclass(obj):
        if mode == 'raw': return dataclasses.asdict(obj)
        obj_class_name = register_object_class(obj)
        return {
            "__type__": "dataclass",
            "__class__": obj_class_name,
            "value": dataclasses.asdict(obj),
        }

    if hasattr(obj, 'as_posix'):
        if mode == 'raw': return obj.as_posix()
        obj_class_name = register_object_class(obj)
        return {
            "__type__": "path",
            "__class__": obj_class_name,
            "value": obj.as_posix(),
        }

    if isinstance(obj, (bytes, bytearray)):
        if mode == 'raw': return obj
        return {
            "__type__": "bytes",
            "value": obj.hex(),
        }

    if isinstance(obj, (set, frozenset)):
        if mode == 'raw': return list(obj)
        return {
            "__type__": "set",
            "value": list(obj),
        }

    if isinstance(obj, Enum):
        if mode == 'raw': return obj.value
        obj_class_name = register_object_class(obj)
        return {
            "__type__": "enum",
            "__class__": obj_class_name,
            "value": obj.value,
        }

    if isinstance(obj, UUID):
        if mode == 'raw': return str(obj)
        return {
            "__type__": "uuid",
            "value": str(obj),
        }

    if isinstance(obj, abc.ABC):
        logger.info(f'Pickle Serializing ABC Object: |r|({type(obj)}) {str(obj)[:1000]}', colored = True)
        # if mode == 'raw': raise TypeError(f"Cannot serialize object of type in raw mode {type(obj)}")
        obj_bytes = default_pickle.dumps(obj)
        if mode == 'raw': return obj_bytes
        return {
            "__type__": "pickle",
            "value": obj_bytes.hex(),
        }


    if hasattr(obj, "numpy"):  # Checks for TF tensors without needing the import
        if mode == 'raw': return obj.numpy().tolist()
        return {
            "__type__": "tensor",
            "value": obj.numpy().tolist(),
        }

    if hasattr(obj, 'tolist'): # Checks for torch tensors without importing
        if mode == 'raw': return obj.tolist()
        return {
            "__type__": "tensor",
            "value": obj.tolist(),
        }

    # Try one shot encoding objects
    # with contextlib.suppress(Exception):

    try:
        logger.info(f'Pickle Serializing Object: |r|({type(obj)}) {str(obj)[:1000]}', colored = True)
        obj_bytes = default_pickle.dumps(obj)
        if mode == 'raw': return obj_bytes
        return {
            "__type__": "pickle",
            "value": obj_bytes.hex(),
        }
    except Exception as e:

        logger.info(f'Error Serializing Object: |r|({type(obj)}) {e}|e| {str(obj)[:1000]}', colored = True)

    raise TypeError(f"Cannot serialize object of type {type(obj)}")


def deserialize_object(
    obj: Union[Dict[str, Any], List[Dict[str, Any]], Any], 
    schema_map: Optional[Dict[str, str]] = None, 
    allow_failed_import: Optional[bool] = False
) -> SerializableObject:
    # sourcery skip: extract-duplicate-method, low-code-quality
    """
    Deserialize an object

    Args:
        obj: the object to deserialize
    """
    if obj is None: return None
    if isinstance(obj, (list, tuple)):
        return [deserialize_object(item, schema_map = schema_map, allow_failed_import = allow_failed_import) for item in obj]

    if isinstance(obj, dict):
        if "__type__" not in obj:
            return {key: deserialize_object(value, schema_map = schema_map, allow_failed_import = allow_failed_import) for key, value in obj.items()}
        
        obj_type = obj["__type__"]
        obj_value = obj["value"]
        if '__class__' in obj:
            if schema_map is not None and obj['__class__'] in schema_map:
                obj['__class__'] = schema_map[obj['__class__']]
            elif obj['__class__'] in _alias_schema_mapping:
                obj['__class__'] = _alias_schema_mapping[obj['__class__']]
            
        if obj_type == "pydantic":
            obj_class_type = obj["__class__"]
            try:
                obj_class = get_object_class(obj_class_type)
                # for k,v in obj_value.items():
                #     if not is_primitive(v):
                #         obj_value[k] = deserialize_object(v)
                return obj_class(**obj_value)
            except ImportError as e:
                if allow_failed_import:
                    return deserialize_object(obj_value, schema_map = schema_map, allow_failed_import = allow_failed_import)
                raise e
        
        if obj_type == "numpy" and np is not None:
            dtype = obj.get("__class__")
            if dtype: dtype = dtype.replace("numpy.", "")
            return np.array(obj_value, dtype = dtype)
        
        if obj_type == "pickle":
            try:
                obj_value = bytes.fromhex(obj_value)
                return default_pickle.loads(obj_value)
            except Exception as e:
                raise TypeError(f"Cannot deserialize object of type {obj_type}: {e}") from e
        
        if obj_type == "datetime":
            return datetime.datetime.fromisoformat(obj_value)
        
        if obj_type == "timedelta":
            return datetime.timedelta(seconds=obj_value)
        
        if obj_type == "dataclass":
            obj_class_type = obj["__class__"]
            # if schema_map is not None and obj_class_type in schema_map:
            #     obj_class_type = schema_map[obj_class_type]
            
            obj_class = get_object_class(obj_class_type)
            return obj_class(**obj_value)
        
        if obj_type == "path":
            obj_class = get_object_class(obj["__class__"])
            return obj_class(obj_value)
        
        if obj_type == "enum":
            obj_class = get_object_class(obj["__class__"])
            return obj_class(obj_value)
        
        if obj_type == "uuid":
            return UUID(obj_value)

        if obj_type == "bytes":
            return bytes.fromhex(obj_value)
        
        if obj_type == "set":
            return set(obj_value)
        
        raise TypeError(f"Cannot deserialize object of type {obj_type}")
    
    if isinstance(obj, bytes):
        # Try to deserialize with pickle
        with contextlib.suppress(Exception):
            return default_pickle.loads(obj)

    return obj






from __future__ import annotations

"""
Qdrant Client with Unified Async / Sync 
"""

import abc
import typing as t
from lzl import load
from lzo.types import BaseModel, eproperty, Field
from lzl.logging import logger, null_logger, Logger
from .config import QdrantClientSettings, fastembed, settings as _settings

if t.TYPE_CHECKING:
    import qdrant_client
    from qdrant_client import grpc as grpc
    from qdrant_client.conversions import common_types as ct
    from qdrant_client.http import ApiClient, SyncApis
    from qdrant_client.http import AsyncApiClient, AsyncApis
    from qdrant_client import QdrantClient as SyncQdrantClient, AsyncQdrantClient
    from fastembed.common import OnnxProvider
else:
    qdrant_client = load.LazyLoad('qdrant_client', 'qdrant_client[fastembed]', install_missing = True)

class QdrantSharedConfig(BaseModel):
    """
    The sharded config for Qdrant
    """
    location: t.Optional[str] = None
    url: t.Optional[str] = None
    port: t.Optional[int] = None
    grpc_port: t.Optional[int] = None
    prefer_grpc: t.Optional[bool] = None
    https: t.Optional[bool] = None
    api_key: t.Optional[str] = None
    prefix: t.Optional[str] = None
    timeout: t.Optional[int] = None
    host: t.Optional[str] = None
    path: t.Optional[str] = None
    
    force_disable_check_same_thread: t.Optional[bool] = None
    auth_token_provider: t.Optional[
        t.Union[t.Callable[[], str], t.Callable[[], t.Awaitable[str]]]
    ] = None
    cloud_inference: t.Optional[bool] = None
    check_compatibility: t.Optional[bool] = None
    kwargs: t.Optional[t.Dict[str, t.Any]] = None

    set_model: t.Optional[str] = Field(None, exclude = True)
    set_model_config: t.Optional[t.Dict[str, t.Any]] = Field(default_factory=dict, exclude = True)
    set_sparse_model: t.Optional[str] = Field(None, exclude = True)
    set_sparse_model_config: t.Optional[t.Dict[str, t.Any]] = Field(default_factory=dict, exclude = True)


    @classmethod
    def build(
        cls, 
        c: 'QdrantClient',

        # These are shard with the config
        url: t.Optional[str] = None,
        port: t.Optional[int] = None,
        grpc_port: t.Optional[int] = None,
        prefer_grpc: t.Optional[bool] = None,
        https: t.Optional[bool] = None,
        api_key: t.Optional[str] = None,
        prefix: t.Optional[str] = None,
        timeout: t.Optional[int] = None,
        host: t.Optional[str] = None,
        path: t.Optional[str] = None,

        # These are from the client
        location: t.Optional[str] = None,
        force_disable_check_same_thread: t.Optional[bool] = None,
        grpc_options: t.Optional[t.Dict[str, t.Any]] = None,
        auth_token_provider: t.Optional[
            t.Union[t.Callable[[], str], t.Callable[[], t.Awaitable[str]]]
        ] = None,
        cloud_inference: t.Optional[bool] = None,
        check_compatibility: t.Optional[bool] = None,
        set_model: t.Optional[str] = None,
        set_sparse_model: t.Optional[str] = None,
        **kwargs: t.Any,
    ) -> 'QdrantSharedConfig':
        """
        Constructs a shared client config
        """
        new = {
            'url': url if url is not None else c.settings.url,
            'port': port or c.settings.port,
            'grpc_port': grpc_port or c.settings.grpc_port,
            'prefer_grpc': prefer_grpc if prefer_grpc is not None else c.settings.prefer_grpc,
            'https': https if https is not None else c.settings.https,
            'api_key': api_key if api_key is not None else c.settings.api_key,
            'prefix': prefix if prefix is not None else c.settings.prefix,
            'timeout': timeout if timeout is not None else c.settings.timeout,
            'host': host if host is not None else c.settings.host,
            'path': path if path is not None else c.settings.path,
            'location': location,
            'force_disable_check_same_thread': force_disable_check_same_thread,
            'grpc_options': grpc_options,
            'auth_token_provider': auth_token_provider,
            'cloud_inference': cloud_inference,
            'check_compatibility': check_compatibility,
            'set_model': set_model,
            'set_sparse_model': set_sparse_model,
            'kwargs': kwargs,
        }
        new = {k:v for k,v in new.items() if v is not None}
        return cls.model_validate(new, context = {'source': 'build', 'client': c})


ResponseT = t.TypeVar('ResponseT')

class QdrantClient(abc.ABC):
    """
    Entry point to communicate with Qdrant service via REST or gRPC API.

    It combines interface classes and endpoint implementation.
    Additionally, it provides custom implementations for frequently used methods like initial collection upload.

    All methods in QdrantClient accept both gRPC and REST structures as an input.
    Conversion will be performed automatically.

    .. note::
        This module methods are wrappers around generated client code for gRPC and REST methods.
        If you need lower-level access to generated clients, use following properties:

        - :py:attr:`QdrantClient.grpc_points`
        - :py:attr:`QdrantClient.grpc_collections`
        - :py:attr:`QdrantClient.rest`

    Args:
        location:
            If `":memory:"` - use in-memory Qdrant instance.
            If `str` - use it as a `url` parameter.
            If `None` - use default values for `host` and `port`.
        url: either host or str of "Optional[scheme], host, Optional[port], Optional[prefix]".
            Default: `None`
        port: Port of the REST API interface. Default: 6333
        grpc_port: Port of the gRPC interface. Default: 6334
        prefer_grpc: If `true` - use gPRC interface whenever possible in custom methods.
        https: If `true` - use HTTPS(SSL) protocol. Default: `None`
        api_key: API key for authentication in Qdrant Cloud. Default: `None`
        prefix:
            If not `None` - add `prefix` to the REST URL path.
            Example: `service/v1` will result in `http://localhost:6333/service/v1/{qdrant-endpoint}` for REST API.
            Default: `None`
        timeout:
            Timeout for REST and gRPC API requests.
            Default: 5 seconds for REST and unlimited for gRPC
        host: Host name of Qdrant service. If url and host are None, set to 'localhost'.
            Default: `None`
        path: Persistence path for QdrantLocal. Default: `None`
        force_disable_check_same_thread:
            For QdrantLocal, force disable check_same_thread. Default: `False`
            Only use this if you can guarantee that you can resolve the thread safety outside QdrantClient.
        auth_token_provider: Callback function to get Bearer access token. If given, the function will be called before each request to get the token.
        check_compatibility: If `true` - check compatibility with the server version. Default: `true`
        set_model: Name of the model to use for encoding documents and queries. Default: `None`
        set_sparse_model: Name of the model to use for hybrid search over documents in combination with dense embeddings. Default: `None`
        **kwargs: Additional arguments passed directly into REST client initialization
    """

    def __init__(
        self,
        location: t.Optional[str] = None,
        url: t.Optional[str] = None,
        port: t.Optional[int] = None,
        grpc_port: t.Optional[int] = None,
        prefer_grpc: t.Optional[bool] = None,
        https: t.Optional[bool] = None,
        api_key: t.Optional[str] = None,
        prefix: t.Optional[str] = None,
        timeout: t.Optional[int] = None,
        host: t.Optional[str] = None,
        path: t.Optional[str] = None,
        force_disable_check_same_thread: bool = False,
        grpc_options: t.Optional[t.Dict[str, t.Any]] = None,
        auth_token_provider: t.Optional[
            t.Union[t.Callable[[], str], t.Callable[[], t.Awaitable[str]]]
        ] = None,
        cloud_inference: bool = False,
        check_compatibility: bool = True,
        set_model: t.Optional[str] = None,
        set_sparse_model: t.Optional[str] = None,
        **kwargs: t.Any,
    ):
        self._extra: t.Dict[str, t.Any] = {}
        self.settings: QdrantClientSettings = _settings
        self.shared_config = QdrantSharedConfig.build(
            self,
            location = location,
            url = url,
            port = port,
            grpc_port = grpc_port,
            prefer_grpc = prefer_grpc,
            https = https,
            api_key = api_key,
            prefix = prefix,
            timeout = timeout,
            host = host,
            path = path,
            force_disable_check_same_thread = force_disable_check_same_thread,
            grpc_options = grpc_options,
            auth_token_provider = auth_token_provider,
            cloud_inference = cloud_inference,
            check_compatibility = check_compatibility,
            set_model = set_model,
            set_sparse_model = set_sparse_model,
            **kwargs,
        )

    """
    Implement interfaces
    """

    @eproperty
    def api(self) -> 'AsyncQdrantClient':
        """
        [Async] Qdrant API client
        """
        _api = qdrant_client.AsyncQdrantClient(
            **self.shared_config.model_dump(exclude_none = True, exclude = {'kwargs'}),
            **self.shared_config.kwargs, 
        )
        if self.shared_config.set_model is not None:
            _api.set_model(self.shared_config.set_model, **self.shared_config.set_model_config)
        if self.shared_config.set_sparse_model is not None:
            _api.set_sparse_model(self.shared_config.set_sparse_model, **self.shared_config.set_sparse_model_config)
        return _api
    
    @property
    def _api_loaded(self) -> bool:
        """
        Returns True if the API is loaded
        """
        return 'api' in self._extra

    @eproperty
    def sapi(self) -> 'SyncQdrantClient':
        """
        [Sync] Qdrant API client
        """
        _api = qdrant_client.QdrantClient(
            **self.shared_config.model_dump(exclude_none = True, exclude = {'kwargs'}),
            **self.shared_config.kwargs, 
        )
        if self.shared_config.set_model is not None:
            _api.set_model(self.shared_config.set_model, **self.shared_config.set_model_config)
        if self.shared_config.set_sparse_model is not None:
            _api.set_sparse_model(self.shared_config.set_sparse_model, **self.shared_config.set_sparse_model_config)
        return _api

    @property
    def _sapi_loaded(self) -> bool:
        """
        Returns True if the SAPI is loaded
        """
        return 'sapi' in self._extra
    
    @t.overload
    def set_model(
        self, 
        embedding_model_name: str,
        max_length: t.Optional[int] = None,
        cache_dir: t.Optional[str] = None,
        threads: t.Optional[int] = None,
        providers: t.Optional[t.Sequence["OnnxProvider"]] = None,
        cuda: bool = False,
        device_ids: t.Optional[t.List[int]] = None,
        lazy_load: bool = False,
        **kwargs: t.Any,
    ) -> None:
        """
        Set embedding model to use for encoding documents and queries.

        Args:
            embedding_model_name: One of the supported embedding models. See `SUPPORTED_EMBEDDING_MODELS` for details.
            max_length (int, optional): Deprecated. Defaults to None.
            cache_dir (str, optional): The path to the cache directory.
                Can be set using the `FASTEMBED_CACHE_PATH` env variable.
                Defaults to `fastembed_cache` in the system's temp directory.
            threads (int, optional): The number of threads single onnxruntime session can use. Defaults to None.
            providers: The list of onnx providers (with or without options) to use. Defaults to None.
                Example configuration:
                https://onnxruntime.ai/docs/execution-providers/CUDA-ExecutionProvider.html#configuration-options
            cuda (bool, optional): Whether to use cuda for inference. Mutually exclusive with `providers`
                Defaults to False.
            device_ids (Optional[list[int]], optional): The list of device ids to use for data parallel processing in
                workers. Should be used with `cuda=True`, mutually exclusive with `providers`. Defaults to None.
            lazy_load (bool, optional): Whether to load the model during class initialization or on demand.
                Should be set to True when using multiple-gpu and parallel encoding. Defaults to False.
        Raises:
            ValueError: If embedding model is not supported.
            ImportError: If fastembed is not installed.
        """
        ...


    def set_model(
        self, 
        embedding_model_name: str,
        **kwargs: t.Any,
    ) -> None:
        self.shared_config.set_model = embedding_model_name
        self.shared_config.set_model_config.update(kwargs)
        if self._api_loaded:
            self.api.set_model(embedding_model_name, **kwargs)
        if self._sapi_loaded:
            self.sapi.set_model(embedding_model_name, **kwargs)
    
    @t.overload
    def set_sparse_model(
        self, 
        embedding_model_name: t.Optional[str] = None,
        cache_dir: t.Optional[str] = None,
        threads: t.Optional[int] = None,
        providers: t.Optional[t.Sequence["OnnxProvider"]] = None,
        cuda: bool = False,
        device_ids: t.Optional[t.List[int]] = None,
        lazy_load: bool = False,
        **kwargs: t.Any,
    ) -> None:
        """
        Set sparse embedding model to use for hybrid search over documents in combination with dense embeddings.

        Args:
            embedding_model_name: One of the supported sparse embedding models. See `SUPPORTED_SPARSE_EMBEDDING_MODELS` for details.
                        If None, sparse embeddings will not be used.
            cache_dir (str, optional): The path to the cache directory.
                                       Can be set using the `FASTEMBED_CACHE_PATH` env variable.
                                       Defaults to `fastembed_cache` in the system's temp directory.
            threads (int, optional): The number of threads single onnxruntime session can use. Defaults to None.
            providers: The list of onnx providers (with or without options) to use. Defaults to None.
                Example configuration:
                https://onnxruntime.ai/docs/execution-providers/CUDA-ExecutionProvider.html#configuration-options
            cuda (bool, optional): Whether to use cuda for inference. Mutually exclusive with `providers`
                Defaults to False.
            device_ids (Optional[list[int]], optional): The list of device ids to use for data parallel processing in
                workers. Should be used with `cuda=True`, mutually exclusive with `providers`. Defaults to None.
            lazy_load (bool, optional): Whether to load the model during class initialization or on demand.
                Should be set to True when using multiple-gpu and parallel encoding. Defaults to False.
        Raises:
            ValueError: If embedding model is not supported.
            ImportError: If fastembed is not installed.

        Returns:
            None
        """
        ...


    def set_sparse_model(
        self, 
        embedding_model_name: t.Optional[str] = None,
        **kwargs: t.Any,
    ) -> None:
        if embedding_model_name is not None:
            self.shared_config.set_sparse_model = embedding_model_name
            self.shared_config.set_sparse_model_config.update(kwargs)
            if self._api_loaded:
                self.api.set_sparse_model(embedding_model_name, **kwargs)
            if self._sapi_loaded:
                self.sapi.set_sparse_model(embedding_model_name, **kwargs)
        


    def close(self, grpc_grace: t.Optional[float] = None, **kwargs: t.Any) -> None:
        """Closes the connection to Qdrant

        Args:
            grpc_grace: Grace period for gRPC connection close. Default: None
        """
        if 'sapi' in self._extra:
            self.sapi.close(grpc_grace = grpc_grace, **kwargs)
            del self._extra['sapi']
        
    async def aclose(self, grpc_grace: t.Optional[float] = None, **kwargs: t.Any) -> None:
        """Closes the connection to Qdrant

        Args:
            grpc_grace: Grace period for gRPC connection close. Default: None
        """
        if 'api' in self._extra:
            await self.api.close(grpc_grace = grpc_grace, **kwargs)
            del self._extra['api']
        self.close(grpc_grace = grpc_grace, **kwargs)


    """
    Sync Interfaces
    """

    @property
    def grpc_collections(self) -> 'grpc.CollectionsStub':
        """gRPC client for collections methods

        Returns:
            An instance of raw gRPC client, generated from Protobuf
        """
        return self.sapi.grpc_collections

    @property
    def grpc_points(self) -> 'grpc.PointsStub':
        """gRPC client for points methods

        Returns:
            An instance of raw gRPC client, generated from Protobuf
        """
        return self.sapi.grpc_points

    @property
    def rest(self) -> 'SyncApis[ApiClient]':
        """REST Client

        Returns:
            An instance of raw REST API client, generated from OpenAPI schema
        """
        return self.sapi.rest
    

    @property
    def http(self) -> 'SyncApis[ApiClient]':
        """REST Client

        Returns:
            An instance of raw REST API client, generated from OpenAPI schema
        """
        return self.sapi.http

    @property
    def init_options(self) -> t.Dict[str, t.Any]:
        """`__init__` Options

        Returns:
             A dictionary of options the client class was instantiated with
        """
        return self.sapi._init_options
    
    """
    Async Interfaces
    """

    @property
    def agrpc_collections(self) -> 'grpc.CollectionsStub':
        """gRPC client for collections methods

        Returns:
            An instance of raw gRPC client, generated from Protobuf
        """
        return self.api.grpc_collections

    @property
    def agrpc_points(self) -> 'grpc.PointsStub':
        """gRPC client for points methods

        Returns:
            An instance of raw gRPC client, generated from Protobuf
        """
        return self.api.grpc_points

    @property
    def arest(self) -> 'AsyncApis[AsyncApiClient]':
        """REST Client

        Returns:
            An instance of raw REST API client, generated from OpenAPI schema
        """
        return self.api.rest

    @property
    def ahttp(self) -> 'AsyncApis[AsyncApiClient]':
        """REST Client

        Returns:
            An instance of raw REST API client, generated from OpenAPI schema
        """
        return self.api.http

    @property
    def ainit_options(self) -> t.Dict[str, t.Any]:
        """`__init__` Options

        Returns:
             A dictionary of options the client class was instantiated with
        """
        return self.api._init_options
    

    

    # we'll probably want to do the same method as kvdb here later.

    """
    Class Wrap Methods
    """

    def _fx_(self, *args, _fx_: t.Optional[str] = None, **kwargs) -> ResponseT:
        """
        [Sync] Wraps the client function
        """
        return getattr(self.sapi, _fx_)(*args, **kwargs)
    
    def _afx_(self, *args, _fx_: t.Optional[str] = None, **kwargs) -> t.Awaitable[ResponseT]:
        """
        [Async] Wraps the client function
        """
        return getattr(self.api, _fx_)(*args, **kwargs)

    @classmethod
    def __init_client_fx__(cls):
        """
        Initializes the class methods
        and sets them based on both the async and sync methods
        """
        import functools
        import inspect
        from makefun import create_function

        excluded_attrs = {
            'DEFAULT_EMBEDDING_MODEL',
            'embedding_models',
            'embedding_model_name',
            'sparse_embedding_models',
            'sparse_embedding_model_name',
            'late_interaction_embedding_models',
            'image_embedding_models',
            'FASTEMBED_INSTALLED',
        }
        existing_methods = set(dir(cls))
        existing_methods.update(excluded_attrs)

        added_methods = set()
        
        # Sync Methods
        for name in dir(qdrant_client.QdrantClient):
            # print(f"Checking method: {name}")
            if name.startswith('_'): continue
            if name in existing_methods: continue
            # print(f"[Sync] Checking method: {name}")
            # if name in skip_methods: continue
            existing_func = getattr(qdrant_client.QdrantClient, name)
            # if not c
            # if not callable(existing_func): continue
            existing_sig = inspect.signature(existing_func)
            new_func = create_function(
                existing_sig,
                functools.partial(cls._fx_, _fx_ = name),
                func_name = name,
                module_name = cls.__module__,
            )
            setattr(cls, name, new_func)
            existing_methods.add(name)
            added_methods.add(name)
            # print(f"[Sync] Added method: {name}")

        # Async Methods
        for name in dir(qdrant_client.AsyncQdrantClient):
            # print(f"Checking async method: {name}")
            if name.startswith('_'): continue
            aname = f'a{name}'
            if aname in {
                'async', 'await'
            }: aname = f'{aname}_' 
            if aname in existing_methods or name in excluded_attrs: continue
            # print(f"[Async] Checking async method: {name}")
            existing_func = getattr(qdrant_client.AsyncQdrantClient, name)
            # if not callable(existing_func): continue
            existing_sig = inspect.signature(existing_func)
            try:
                new_func = create_function(
                    existing_sig,
                    functools.partial(cls._afx_, _fx_ = name),
                    func_name = aname,
                    module_name = cls.__module__,
                )
                setattr(cls, aname, new_func)
                existing_methods.add(aname)
                added_methods.add(aname)
                # print(f"[Async] Added async method: {name} -> {aname}")
            except Exception as e:
                # print(f"Error adding method: {name} -> {aname}")
                raise e
        
# stubgen --inspect-mode --include-private --include-docstrings client_v2.py 
# stubgen --inspect-mode --include-docstrings --include-private --ignore-errors -m lzl.api.qdrant.client
QdrantClient.__init_client_fx__()



import random
import inspect
import bisect
import itertools
from lzl.api import aiohttpx
from datetime import datetime, timedelta
from typing import Dict, Optional, Iterator, AsyncIterator, Union, List, Tuple
from lzo.utils import timer, is_coro_func

__all__ = [
    'is_naive',
    'total_seconds',
    'remove_trailing_slash',
    'full_name',
    'merge_dicts',
    'is_coro_func',
    # 'timed',
    'timer',
    'parse_stream',
    'aparse_stream',
]


def merge_dicts(x: Dict, y: Dict):
    z = x.copy()
    z.update(y)
    return z


def full_name(func, follow_wrapper_chains=True):
    """
    Return full name of `func` by adding the module and function name.

    If this function is decorated, attempt to unwrap it till the original function to use that
    function name by setting `follow_wrapper_chains` to True.
    """
    if follow_wrapper_chains: func = inspect.unwrap(func)
    return f'{func.__module__}.{func.__qualname__}'

def is_naive(dt: datetime):
    """Determines if a given datetime.datetime is naive."""
    return dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None


def total_seconds(delta: timedelta):
    """Determines total seconds with python < 2.7 compat."""
    # http://stackoverflow.com/questions/3694835/python-2-6-5-divide-timedelta-with-timedelta
    return (delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1e6) / 1e6


def remove_trailing_slash(host: str):
    """
    Removes trailing slash from a host if it exists.
    """
    return host[:-1] if host.endswith("/") else host


def parse_stream_line_bytes(line: bytes) -> Optional[str]:
    """
    Parse a line from a Server-Sent Events stream.
    """
    if line:
        if line.strip() == b"data: [DONE]":
            # return here will cause GeneratorExit exception in urllib3
            # and it will close http connection with TCP Reset
            return None
        if line.startswith(b"data: "):
            line = line[len(b"data: "):]
            return line.decode("utf-8")
        else:
            return None
    return None


def parse_stream_line_string(line: str) -> Optional[str]:
    """
    Parse a line from a Server-Sent Events stream.
    """
    if line:
        if line.strip() == "data: [DONE]":
            # return here will cause GeneratorExit exception in urllib3
            # and it will close http connection with TCP Reset
            return None
        return line[len("data: "):] if line.startswith("data: ") else None
    return None

def parse_stream_line(line: Union[str, bytes]) -> Optional[str]:
    """
    Parse a line from a Server-Sent Events stream.
    """
    if isinstance(line, bytes):
        return parse_stream_line_bytes(line)
    elif isinstance(line, str):
        return parse_stream_line_string(line)
    else:
        raise TypeError("line must be str or bytes")


def parse_stream(response: aiohttpx.Response) -> Iterator[str]:
    """
    Parse a Server-Sent Events stream.
    """
    for line in response.iter_lines():
        _line = parse_stream_line(line)
        if _line is not None:
            yield _line

async def aparse_stream(response: aiohttpx.Response) -> AsyncIterator[str]:
    """
    Parse a Server-Sent Events stream.
    """
    async for line in response.aiter_lines():
        _line = parse_stream_line(line)
        if _line is not None:
            yield _line


def weighted_choice(choices: Union[List[Tuple[str, float]], Dict[str, float]]) -> str:
   """
   Randomly selects a choice based on the weights provided
   """
   if isinstance(choices, dict):
       choices = list(choices.items())
   weights = list(zip(*choices))[1]
   return choices[bisect.bisect(
       list(itertools.accumulate(weights)),
        random.uniform(0, sum(weights))
    )][0]

